from __future__ import annotations

import time
import unittest

import httpx
from azul_bedrock import dispatcher

from . import plugin_support as sup
from . import mock_dispatcher as md


class TestMockFileServer(unittest.TestCase):
    def setUp(self) -> None:
        md.clear()
        return super().setUp()

    @classmethod
    def setUpClass(cls) -> None:
        cls.mock_server, cls.server = sup.setup_mock_dispatcher()
        cls.dp = dispatcher.DispatcherAPI(
            events_url=cls.server,
            data_url=cls.server + "/mock",
            retry_count=2,
            timeout=2,
            author_name="TestStorageProxyFile",
            author_version="1",
            deployment_key="scorpion",
        )
        cls.editor = md.Editor(cls.server)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.mock_server.stop()
        cls.mock_server.kill()

    def test_get_zero_file(self):
        self.editor.set_stream("test-zeros-100", 200, b"\x00" * 100)
        resp = httpx.get(self.server + "/api/v3/stream/source/content/test-zeros-100", headers={"range": "bytes=0-50"})
        self.assertEqual(resp.status_code, 206, resp.content)
        self.assertEqual(resp.content, b"\x00" * 51)
        self.assertEqual(resp.headers["content-range"], "bytes 0-50/100")
        last = self.editor.get_last_request()
        self.assertEqual(last.range_raw, "bytes=0-50")
        self.assertEqual(last.range, [0, 50])

        self.editor.set_stream("test-zeros-50", 200, b"\x00" * 50)
        resp = httpx.get(self.server + "/api/v3/stream/source/content/test-zeros-50", headers={"range": "bytes=-30"})
        self.assertEqual(resp.status_code, 206, resp.content)
        self.assertEqual(resp.content, b"\x00" * 30)
        self.assertEqual(resp.headers["content-range"], "bytes 20-49/50")
        last = self.editor.get_last_request()
        self.assertEqual(last.range_raw, "bytes=-30")
        self.assertEqual(last.range, [20, 49])

        self.editor.set_stream("test-zeros-20", 200, b"\x00" * 20)
        resp = httpx.get(self.server + "/api/v3/stream/source/content/test-zeros-20", headers={"range": "bytes=-50"})
        self.assertEqual(resp.status_code, 206, resp.content)
        self.assertEqual(resp.content, b"\x00" * 20)
        self.assertEqual(resp.headers["content-range"], "bytes 0-19/20")
        last = self.editor.get_last_request()
        self.assertEqual(last.range_raw, "bytes=-50")
        self.assertEqual(last.range, [0, 19])

        self.editor.set_stream("test-zeros-100", 200, b"\x00" * 100)
        resp = httpx.get(
            self.server + "/api/v3/stream/source/content/test-zeros-100", headers={"range": "bytes=50-500"}
        )
        self.assertEqual(resp.status_code, 206, resp.content)
        self.assertEqual(resp.content, b"\x00" * 50)
        self.assertEqual(resp.headers["content-range"], "bytes 50-99/100")
        last = self.editor.get_last_request()
        self.assertEqual(last.range_raw, "bytes=50-500")
        self.assertEqual(last.range, [50, 99])

        self.editor.set_stream("test-zeros-256", 200, b"\x00" * 256)
        resp = httpx.get(self.server + "/api/v3/stream/source/content/test-zeros-256", headers={"range": "bytes=256-"})
        self.assertEqual(resp.status_code, 416, resp.content)
        self.assertEqual(resp.content, b"")
        self.assertEqual(resp.headers["content-range"], "bytes */256")
        last = self.editor.get_last_request()
        self.assertEqual(last.range_raw, "bytes=256-")
        self.assertEqual(last.range, None)

        self.editor.set_stream("test-zeros-100", 200, b"\x00" * 100)
        resp = httpx.get(self.server + "/api/v3/stream/source/content/test-zeros-100", headers={"range": "bytes=500-"})
        self.assertEqual(resp.status_code, 416, resp.content)
        self.assertEqual(resp.content, b"")
        self.assertEqual(resp.headers["content-range"], "bytes */100")
        last = self.editor.get_last_request()
        self.assertEqual(last.range_raw, "bytes=500-")
        self.assertEqual(last.range, None)

    def test_get_ff_file(self):
        self.editor.set_stream("test-ff-200", 200, b"\xff" * 100)
        resp = httpx.get(self.server + "/api/v3/stream/source/content/test-ff-200", headers={"range": "bytes=0-50"})
        self.assertEqual(resp.status_code, 206, resp.content)
        self.assertEqual(resp.content, b"\xff" * 51, len(resp.content))
        last = self.editor.get_last_request()
        self.assertEqual(last.range_raw, "bytes=0-50")
        self.assertEqual(last.range, [0, 50])

    def test_get_mod256(self):
        self.editor.set_stream("test-mod256", 200, bytes(range(256)))
        resp = httpx.get(self.server + "/api/v3/stream/source/content/test-mod256")
        self.assertEqual(resp.status_code, 200, resp.content)
        last = self.editor.get_last_request()
        self.assertEqual(last.range, None)
        self.assertEqual(len(resp.content), 256)

        self.editor.set_stream("test-mod256", 200, bytes(range(256)))
        resp = httpx.get(self.server + "/api/v3/stream/source/content/test-mod256", headers={"range": "bytes=50-500"})
        self.assertEqual(resp.status_code, 206, resp.content)
        self.assertEqual(resp.headers["content-range"], "bytes 50-255/256")
        last = self.editor.get_last_request()
        self.assertEqual(last.range, [50, 255])
        self.assertEqual(len(resp.content), 206)

        self.editor.set_stream("test-mod256", 200, bytes(range(256)))
        resp = httpx.get(self.server + "/api/v3/stream/source/content/test-mod256", headers={"range": "bytes=-2"})
        self.assertEqual(resp.status_code, 206, resp.content)
        self.assertEqual(resp.headers["content-range"], "bytes 254-255/256")
        last = self.editor.get_last_request()
        self.assertEqual(last.range, [254, 255])
        self.assertEqual(len(resp.content), 2)



"""Transform models to/from network event format and runner internal format."""

import copy
import datetime
import json
import logging

from azul_bedrock import models_network as azm
from pydantic import TypeAdapter

from .models import Event, EventData, EventParent, Feature, FeatureValue, JobResult
from .plugin import Multiplugin, Plugin

logger = logging.getLogger(__name__)


def get_registrations(plugin: Plugin) -> list[azm.PluginEvent]:
    """Generate registration messages for all multiplugins."""
    ret = []
    for multiplugin in plugin.get_registered_multiplugins():
        now = datetime.datetime.now(datetime.timezone.utc)
        # send all config that is not prefixed with 'secret_'
        # encode as json to get the real value to allow future access to it via the registration.
        safe_config = {x: json.dumps(y) for x, y in plugin.cfg.model_dump().items() if not x.startswith("secret_")}
        mp = plugin.get_multiplugin(multiplugin)
        description = plugin.DESCRIPTION
        if mp.description is not None and len(mp.description) > 0:
            description = mp.description

        author_details = azm.PluginEvent.Entity(
            **gen_author(plugin, mp).model_dump(exclude_defaults=True),
            description=description,
            # sort features by name
            features=TypeAdapter(list[Feature]).dump_python(sorted(plugin.FEATURES)),
            contact=plugin.CONTACT,
            # sort config by key
            config=dict(sorted(safe_config.items())),
        )
        reg_event = azm.PluginEvent(
            model_version=azm.CURRENT_MODEL_VERSION,
            kafka_key="runner-placeholder",
            entity=author_details,
            timestamp=now,
            author=author_details.summary(),
        )
        ret.append(reg_event)
    return ret


def gen_author(plugin: Plugin, mp: Multiplugin | None) -> azm.Author:
    """Return an author object for the multiplugin."""
    name = plugin.NAME
    version = plugin.VERSION
    security = plugin.SECURITY
    if mp:
        # add multiplugin name as a suffix
        if mp.name:
            name += f"-{mp.name}"
        # add multiplugin version as a suffix
        if mp.version:
            version += f"-{mp.version}"
        if mp.security:
            security = mp.security

    return azm.Author(
        category="plugin",
        name=name,
        version=version,
        security=security,
    )


def gen_processing_events(
    plugin: Plugin, posted: dict[str, azm.Datastream], author: azm.Author, src: azm.BinaryEvent, result: JobResult
) -> azm.StatusEvent:
    """Generates new output events based on the results of this plugin run."""
    # history/audit is added at every step that produces a message (including source events)
    # entity summary can be useful for retaining context about each step which would otherwise be
    # lost on a message as it decomposed into it's children
    # entity summary is expected to contain 'type', 'size' and optionally 'filename' for binaries
    # history also captures any 'relationship' information for how this message/child was derived
    # from its parent event/entity.  This will currently contain the plugin's relationship info.
    now = _get_now()
    out_events = []
    for event in result.events:
        if not event.parent:
            # event has no parent
            out_events.append(_gen_event_output(plugin, posted, author, src, event, now))
        else:
            out_events.append(_gen_event_extracted(plugin, posted, author, src, event, now))

    # do not track large fields from input event on outgoing status
    # this reduces the size of outgoing event
    min_src = copy.copy(src)  # avoid side effects
    min_src.entity = copy.copy(src.entity)  # avoid side effects
    min_src.entity.features = []
    min_src.entity.datastreams = []
    min_src.entity.info = {}

    # build outgoing status
    return azm.StatusEvent(
        model_version=azm.CURRENT_MODEL_VERSION,
        # provide a temporary id, as dispatcher must calculate this value
        kafka_key="runner-placeholder",
        timestamp=now,
        author=author,
        entity=azm.StatusEvent.Entity(
            input=min_src,
            status=result.state.label,
            error=result.state.failure_name or None,
            message=result.state.message or None,
            results=out_events,
            runtime=result.runtime,
        ),
    )


def _to_api_features(plugin: Plugin, feats_in: dict[str, set[FeatureValue]]) -> list[azm.FeatureValue]:
    """Convert a structured dict of features-by-name to a list of (name, value, <meta>) for the API.

    Any features that are not registered with the plugin are silently dropped,
    they should have been filtered during plugin execution.
    """
    ret = []
    for feature in plugin.FEATURES:
        for v in sorted(feats_in.get(feature.name, [])):
            fv = azm.FeatureValue(
                name=feature.name,
                type=feature.type,
                value=v.value_encoded(),
                label=v.label,
                offset=v.offset,
                size=v.size,
            )
            ret.append(fv)
    return ret


def _to_api_content(posted: dict[str, azm.Datastream], data: list[EventData]) -> list[azm.Datastream]:
    """Convert data from runner to the structure used by the restapi."""
    data = sorted(data)
    # update label and language for posted data
    return [
        posted[x.hash].model_copy(update=dict(label=x.label, language=x.language)) for x in data if x.hash in posted
    ]


def _gen_event_output(
    plugin: Plugin,
    posted: dict[str, azm.Datastream],
    author: azm.Author,
    src: azm.BinaryEvent,
    event: Event,
    now: datetime.datetime,
) -> azm.BinaryEvent:
    """Generate an enrichment event for existing entity."""
    # enriched events drop the entity.data of parent event
    action = azm.BinaryAction.Enriched
    if plugin._IS_USING_PUSHER:
        # Action is the same as whatever the source action is (sourced or mapped are the only two expected)
        action = src.action

    out_data = []
    extra_data = _to_api_content(posted, event.data)
    if extra_data:
        # this plugin run produced extra event data
        action = azm.BinaryAction.Augmented
        # Remove content entries other than the primary content stream and newly generated streams
        out_data += [ds for ds in (src.entity.datastreams or []) if ds.label == "content"]
        out_data += extra_data

    out_event = azm.BinaryEvent(
        model_version=azm.CURRENT_MODEL_VERSION,
        # provide a temporary id, as dispatcher must calculate this value
        kafka_key="runner-placeholder",
        action=action,
        timestamp=now,
        source=src.source.model_copy(
            update=dict(
                path=copy.copy(src.source.path)
                + [
                    azm.PathNode(
                        action=action,
                        timestamp=now,
                        author=author,
                        sha256=src.entity.sha256,
                    )
                ]
            )
        ),
        author=author,
        entity=src.entity.model_copy(
            update=dict(
                features=_to_api_features(plugin, event.features),
                datastreams=out_data,
                info=event.info,
            )
        ),
    )
    if plugin._IS_USING_PUSHER:
        # Pusher doesn't need any additional path added because the provided event is the output event..
        out_event.source.path.pop()
        src_entity_copy = src.entity.model_copy(deep=True)
        # # Merging features
        for base_feats in src_entity_copy.features:
            out_event.entity.features.append(base_feats)

        # Merging datastreams
        for base_stream in src_entity_copy.datastreams:
            is_stream_already_present = False
            for out_stream in out_event.entity.datastreams:
                if base_stream.label == out_stream.label:
                    is_stream_already_present = True
                    break
            # If the stream is already in the output don't add it again.
            if is_stream_already_present:
                continue

            out_event.entity.datastreams.append(base_stream)

        # Add base info if the plugin didn't add any.
        if src_entity_copy.info and not out_event.entity.info:
            out_event.entity.info = src_entity_copy.info
    return out_event


def _gen_event_extracted(
    plugin: Plugin,
    posted: dict[str, azm.Datastream],
    author: azm.Author,
    src: azm.BinaryEvent,
    event: Event,
    now: datetime.datetime,
) -> azm.BinaryEvent:
    """Generate an extraction event for a new entity."""
    # generate path between this event and the original source-entity submission
    path_extension = []
    c = event.as_parent()
    while c.parent:
        try:
            # For 'binary', entity.sha256 is equivalent to the hash of the 'content' stream
            pc: azm.Datastream | EventParent = posted[c.sha256]
        except KeyError:
            # try data-less submission
            pc = c

        # add to front of list
        path_extension.insert(
            0,
            azm.PathNode(
                action=azm.BinaryAction.Extracted,
                timestamp=now,
                author=author,
                relationship=c.relationship,
                sha256=c.sha256,
                file_format=pc.file_format,
                size=pc.size,
                filename=c.filename,
                language=c.language,
            ),
        )
        # now inspect parent, unless it is top parent
        c = c.parent

    # generate output entity
    try:
        # For 'binary', entity.sha256 is equivalent to the hash of the 'content' stream
        pc = posted[event.sha256]
        out_entity: azm.BinaryEvent.Entity = pc.to_input_entity()
    except KeyError:
        # try data-less submission
        out_entity: azm.BinaryEvent.Entity = event.to_input_entity()

    out_entity.sha256 = event.sha256
    out_entity.features = out_entity.features + _to_api_features(plugin, event.features)
    out_entity.datastreams = _to_api_content(posted, event.data)
    out_entity.info = event.info

    return azm.BinaryEvent(
        model_version=azm.CURRENT_MODEL_VERSION,
        # provide a temporary id, as dispatcher must calculate this value
        kafka_key="runner-placeholder",
        action=azm.BinaryAction.Extracted,
        timestamp=now,
        source=src.source.model_copy(update=dict(path=copy.copy(src.source.path) + path_extension)),
        author=author,
        entity=out_entity,
    )


def _get_now():
    """Exists as cannot mock built-in datetime."""
    return datetime.datetime.now(datetime.timezone.utc)



"""Plugin template for handling plugin metadata and execution methods.

Coordination of execution and parsing of results is beyond the scope of this module.
"""

import logging
import typing
from typing import Callable, ClassVar, Iterable, Optional, Type, Union

from azul_bedrock.models_network import FeatureType
from pydantic import BaseModel

from . import settings
from .models import Event, Feature, Job, State

logger = logging.getLogger(__name__)


class Multiplugin(BaseModel):
    """Class to store data about multi-plugins."""

    name: str | None = None
    version: str | None = None
    security: str | None = None
    description: str | None = None
    callback: Callable[[Job], Optional[Union[State, State.Label]]]


class Plugin:
    """Defines the template for an Azul 3 plugin."""

    # Plaintext security string applied to all results of a plugin.
    # Examples:
    # SECURITY = "UNCLASSIFIED"
    # SECURITY = "UNCLASSIFIED TLP:GREEN"
    # SECURITY = "AMAZING TOMATO TLP:GREEN REL:ME,YOU,THATONE"
    SECURITY: ClassVar[Optional[str]] = None

    # Default settings for this plugin. This is accumulated through all descending classes.
    # Child classes will inherit unless they override in their own SETTINGS.
    SETTINGS: ClassVar[Type[settings.PluginBaseSettings]] = settings.PluginBaseSettings

    NAME: ClassVar[str]  # Plugin name, set automatically from class name.
    DESCRIPTION: ClassVar[str]  # Plugin description, set automatically from class comment.
    VERSION: ClassVar[str] = ""  # Plugin version
    CONTACT: ClassVar[Optional[str]] = None  # Contact point for issues with this plugin

    # Toggle output entity type to mapped from if no output streams are provided. (must be enriched or mapped)
    _IS_USING_PUSHER: ClassVar[bool] = False

    # List of features that the plugin produces.
    FEATURES: ClassVar[Iterable[Feature]] = {
        # Standard features set by all binary plugins.
        # Child classes that set this property will ADD to this feature set.
        Feature(name="malformed", desc="File is malformed in some way.", type=FeatureType.String),
        # these features are set from bedrock code as part of generating the binary entity
        Feature(name="file_extension", desc="File extension of the 'content' stream.", type=FeatureType.String),
        Feature(name="file_format", desc="Assemblyline file type of the 'content' stream.", type=FeatureType.String),
        Feature(name="filename", desc="Name on disk of the 'content' stream.", type=FeatureType.Filepath),
        Feature(name="magic", desc="File magic found for the 'content' stream.", type=FeatureType.String),
        Feature(name="mime", desc="Mimetype found for the 'content' stream.", type=FeatureType.String),
    }

    def __init__(self, config: settings.Settings | dict | None = None) -> None:
        if not isinstance(config, settings.Settings):
            config = settings.parse_config(type(self), config or {})
            logger.warning("plugin started using simple config mode, this is intended only for tests")
        config = self._alter_config(config)

        # parse config
        self.cfg = config
        # log settings if they are interesting
        to_print = self.cfg.model_dump(exclude_defaults=True, exclude_unset=True)
        if to_print:
            logger.info("custom startup options:")
            for k, v in self.cfg.model_dump(exclude_defaults=True, exclude_unset=True).items():
                logger.info(f"{k:20}: {v}")

        self.shutting_down = False

        # add suffix to name and version
        self.NAME = f"{type(self).__name__}"
        self.NAME = self.NAME.replace(self.cfg.name_remove_prefix, "", 1)
        if ns := self.cfg.name_suffix:
            self.NAME += f"-{ns}"
        if vs := self.cfg.version_suffix:
            self.VERSION += f"-{vs}"
        # load description from docstring
        desc = type(self).__doc__
        self.DESCRIPTION = desc if desc else ""

        # create logger for plugin creator to use
        self.logger = logging.getLogger("azul.plugin." + self.NAME.lower())

        # add the multiplugin for execute()
        self._multiplugins: dict[str | None, Multiplugin] = {
            None: Multiplugin(name=None, version=None, callback=self._mp_callback, description=None)
        }

        # travel inheritance hierarchy for all defined features
        features = {}
        for c in reversed(type(self).__mro__):
            if hasattr(c, "FEATURES"):
                for f in c.FEATURES:
                    features[f.name] = f
        self.FEATURES = sorted(features.values())

        # override security through config
        self.SECURITY = self.cfg.security_override or self.SECURITY
        if self.SECURITY:
            logger.info(f"running with security: {self.SECURITY}")
        if self.SECURITY and not isinstance(self.SECURITY, str):
            logger.error(f"self.SECURITY is {self.SECURITY}, not a string")
            exit(1)

    def _mp_callback(self, job: Job):
        """Defined method for multiplugin callback."""
        return self.execute(job)

    def _alter_config(self, config: settings.Settings) -> settings.Settings:
        """Implement this function to edit config immediately after config generation.

        This is also a function for editing other metadata of the plugin including FEATURES.
        Helpful for plugins that dynamically figure out features that are needed.
        """
        return config

    def reset(self, job: Job):
        """Clear the current result object."""
        self._job = job
        # binary data generated by plugin
        # hash -> binary, tags
        self.data: dict[str, typing.BinaryIO] = {}
        self.events: list[Event] = []
        # map sha256 -> Event
        self._event_map: dict[str | None, Event] = {}

        # main event always exists
        self._event_main = self.get_data_event(None)
        # make main event easier to manipulate
        self.add_data = self._event_main.add_data
        self.add_data_file = self._event_main.add_data_file
        self.add_info = self._event_main.add_info
        self.add_feature_values = self._event_main.add_feature_values
        self.add_many_feature_values = self._event_main.add_many_feature_values
        self._add_child = self._event_main._add_child
        self.add_child_with_data = self._event_main.add_child_with_data
        self.add_child_with_data_file = self._event_main.add_child_with_data_file
        self.add_text = self._event_main.add_text

    def get_data_event(self, data_hash: str | None):
        """Creates, or retrieve an event for manipulation.

        The data_hash provided is either None or the parent of the original job Event.
        If it's None the event is the root event and has no parent.

        If the data_hash is set it's value is the sha256 of the new event that is a parent of the current sha256.
        This is typically used to link network captures to the current job.
        """
        if data_hash not in self._event_map:
            ret = self._event_map.setdefault(
                data_hash,
                Event(
                    sha256=self._job.event.entity.sha256,
                    parent_sha256=data_hash,
                ),
            )
            ret.set_stores(self.data, self.events)
            self.events.append(ret)
        return self._event_map[data_hash]

    def register_multiplugin(
        self,
        name: str,
        version: str,
        callback: Callable[[Job], Optional[Union[State, State.Label]]],
        *,
        description: str | None = None,
        security: str | None = None,
    ):
        """Register a multiplugin callback for execution as part of the plugin run."""
        self._multiplugins[name] = Multiplugin(
            name=name, version=version, security=security, description=description, callback=callback
        )

    def get_multiplugin(self, name: str | None):
        """Get a multiplugin by it's name."""
        return self._multiplugins.get(name)

    def get_registered_multiplugins(self):
        """Return a list of the available multiplugins."""
        return self._multiplugins.keys()

    def is_ready(self) -> bool:
        """Return whether the plugin is ready to accept jobs.

        Note: It is called before each job fetch attempt which,
        depending on plugin, may have performance consequences.
        """
        return True

    def execute(self, job: Job) -> Optional[Union[State, State.Label]]:
        """Default entrypoint for plugin execution."""
        raise NotImplementedError

    def is_malformed(self, message: str):
        """Plugins should return this when a file has been processed and is invalid compared to expected format.

        Many plugins receive files that look correct for them but turn out to be corrupted in some known way.
        This is an attempt to standardise the way malformed files are handled.
        """
        self.add_feature_values("malformed", message)
        return State(label=State.Label.COMPLETED_WITH_ERRORS, message=message)



"""Handle full Plugin execution loop.

* Receive jobs from external source or local
* Send job results to external source or local
* Run required multiplugins
* Handle timeouts and critical errors
"""

import contextlib
import datetime
import logging
import multiprocessing
import multiprocessing.sharedctypes
import multiprocessing.util
import queue as queueLib
import signal
import subprocess  # noqa: S404 # noqa S404
import sys
import time
import traceback
from typing import Any, Generator, Optional, Type

from azul_bedrock import models_network as azm
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

from . import models, network, plugin_executor, settings
from . import plugin as mplugin
from .models import JobResult, State, TaskModel
from .storage import StorageError, StorageProxyFile

logger = logging.getLogger(__name__)


QUEUE_PUT_TIMEOUT = 0.5


class CriticalError(Exception):
    """Something has gone wrong that indicates the plugin needs to exit."""

    pass


class SigTermExitError(Exception):
    """Sig term has been provided to the process and it is not exiting."""

    pass


class RecreateException(Exception):
    """Plugin needs to be re-created by it's parent process.

    This is raised when the source plugin has detected a file change.
    """

    pass


def get_git_version_suffix(config: settings.Settings) -> str | None:
    """Get the git version suffix for the watched repo if there is one."""
    if config.watch_path and config.watch_type == settings.WatchTypeEnum.GIT:
        # wait for valid git repo on disk
        for _ in range(3):
            retcode = subprocess.call(
                ["git", "status"],  # noqa: S607
                cwd=config.watch_path,
                shell=False,
            )
            if retcode == 0:
                break
            logger.warning(f"'git status' failed with {retcode} on {config.watch_path}, sleeping")
            time.sleep(config.watch_wait)

        # recalculate commit hash as version_suffix
        # requires git to be installed
        try:
            output = subprocess.check_output(
                ["git", "rev-parse", "--short", "HEAD"],  # noqa: S607
                cwd=config.watch_path,
                shell=False,
            )
        except subprocess.CalledProcessError as e:
            raise CriticalError(f"is git not installed or {config.watch_path} not a valid git checkout?") from e
        return output.decode().strip()
    return None


class WatchPath(FileSystemEventHandler):
    """Event handler for watchdog."""

    def __init__(self, watch_wait: int) -> None:
        self._restart_after = None
        self._watch_wait = watch_wait
        super().__init__()

    def on_created(self, event):
        """Event trigger."""
        self._update(event)

    def on_deleted(self, event):
        """Event trigger."""
        self._update(event)

    def on_moved(self, event):
        """Event trigger."""
        self._update(event)

    def on_modified(self, event):
        """Event trigger."""
        self._update(event)

    def _update(self, event: FileSystemEvent):
        """Trigger restart if necessary."""
        logger.info(f"Watched file {event.event_type}: {event.src_path}")
        if not self._restart_after:
            logger.info(f"Plugin will be restarted after at least {self._watch_wait} seconds.")
            # delay in case multiple files are being rewritten
            self._restart_after = time.time() + self._watch_wait

    def check_updated(self) -> bool:
        """Return true if path was updated."""
        if self._restart_after and self._restart_after < time.time():
            self._restart_after = None
            return True
        else:
            return False


class Coordinator:
    """Manage continuous plugin execution and prevent bad jobs from stalling plugin."""

    _plugin: mplugin.Plugin

    def __init__(
        self,
        plugin_class: Type[mplugin.Plugin],
        config: settings.Settings,
    ):
        self._plugin_class = plugin_class
        if isinstance(config, dict):
            raise ValueError("Config should be a Settings object not a dictionary!")
        self._cfg = config

        self.is_signalled_to_exit = False
        signal.signal(signal.SIGINT, self.set_signal_exit)
        signal.signal(signal.SIGTERM, self.set_signal_exit)

        self._watchdog: Any | None = None
        self._watched: WatchPath | None = None

        # start watchdog
        if self._cfg.watch_path:
            self._watchdog = Observer()
            self._watched = WatchPath(self._cfg.watch_wait)
            # sleep before starting plugin
            # In Kubernetes, the sidecar is unlikely to have completed sync yet
            # so this sleep avoids kubernetes pod restarts.
            time.sleep(self._cfg.watch_wait)
            self._watchdog.schedule(self._watched, self._cfg.watch_path, recursive=True)
            self._watchdog.start()
        self._recreate_plugin()

    def set_signal_exit(self, *args):
        """Set the option to exit the plugin based on a signal (SIGINT, SIGTERM...)."""
        self.is_signalled_to_exit = True

    @property
    def plugin(self):
        """Return plugin instance."""
        return self._plugin

    @property
    def cfg(self):
        """Return config instance."""
        return self._cfg

    def __del__(self):
        """Cleanup on deletion."""
        if hasattr(self, "_watchdog") and self._watchdog:
            self._watchdog.stop()

    def _recreate_plugin(self):
        """Recreate plugin including git version suffix."""
        version_suffix = get_git_version_suffix(self._cfg)
        if version_suffix:
            self._cfg.version_suffix = version_suffix
        self._plugin = self._plugin_class(config=self._cfg)
        self._network = network.Network(self._plugin)

    def run_once(
        self,
        event: azm.BinaryEvent,
        local_streams: list[StorageProxyFile] = None,
    ) -> dict[str, JobResult]:
        """Perform a local run of plugin, with timeout and error capture."""
        if self._watched and self._watched.check_updated():
            self._recreate_plugin()
            logger.info("Plugin has been restarted due to a file change")
        # keep newest result for each multiplugin
        results = {}
        for result, multiplugin in self._run_job(event, local=local_streams):
            results[multiplugin] = result
        return results

    def run_loop(
        self,
        *,
        queue: multiprocessing.Queue,
        job_limit: int = None,
    ):
        """Run continually and fetch jobs from the queue for processing while posting results to dispatcher.

        If `job_limit` is given, the loop will terminate after completing that many jobs (whether successful or not).
        Relies on network interactions to handle their own errors. If they fail then the runner will die.
        """
        if not self._cfg.events_url:
            raise ValueError("Cannot run fetch/execute loop when events_url is None")

        self._network.post_registrations()
        job_count = 0
        while job_limit is None or (job_limit and job_limit > job_count):
            if self._watched and self._watched.check_updated():
                logger.info("Plugin being restarted due to a file change")
                raise RecreateException()
            # ensure plugin is (still) ready to receive jobs
            if not self._plugin.is_ready():
                backoff = self._cfg.not_ready_backoff
                logger.info("Plugin is not ready to receive jobs; retrying in %ds" % backoff)
                time.sleep(backoff)
                continue

            # Set current job to None to indicate fetching a job.
            with contextlib.suppress(queueLib.Full):
                queue.put(None, block=False)

            # Exit gracefully when a SIGTERM Or SIGINT is provided. (placed after queue.put for testing reasons)
            if self.is_signalled_to_exit:
                raise SigTermExitError()

            event = self._network.fetch_job()
            # Put task immediately to account for time spend fetching a file.
            try:
                queue.put(TaskModel(in_event=event, start_time_epoch=time.time()), timeout=QUEUE_PUT_TIMEOUT)
            except queueLib.Full:
                logger.warning("monitoring queue is full bypassing, which may result in odd heartbeats.")

            job_count += 1
            try:
                for result, multiplugin in self._run_job(event, queue=queue):
                    self._network.ack_job(event, result, multiplugin)
                    # close any open handles in JobResult
                    result.close()
            except CriticalError:  # No extra logging required.
                logger.info("Exited on a critical error.")
                raise
            except Exception:  # Unknown error so extra logging required.
                traceback.print_exc()
                logger.error(
                    f"Error during job with kafka_key: '{event.kafka_key}', action: '{event.action}',"
                    + f" author info: {event.author.model_dump()}"
                    + f"\n source info: {event.source.model_dump()}."
                )
                raise
        sys.exit(0)

    def _run_job(
        self,
        event: azm.BinaryEvent,
        local: list[StorageProxyFile] | None = None,
        queue: multiprocessing.Queue = None,
    ) -> Generator[tuple[JobResult, str], None, None]:
        """Run plugin with supplied event and yield response.

        This is a generator so that yielded results may be published while plugin continues to run.
        Useful if a certain multiplugin exceeds max memory and crashes the process.
        """
        logger.info(
            "received plugin=%s file_format=%s size=%s sha256=%s"
            % (
                self._plugin.NAME,
                event.entity.file_format,
                event.entity.size,
                event.entity.sha256,
            )
        )
        run_start = datetime.datetime.now(datetime.timezone.utc)

        # check whether the maximum depth limit for the event has been reached, and opt-out if it has been.
        max_depth = self._cfg.plugin_depth_limit
        depth_reached = False
        if max_depth >= 0:
            if len(event.source.path) >= max_depth:
                depth_reached = True

        if depth_reached:
            result = JobResult(
                state=State(
                    State.Label.OPT_OUT,
                    message=f"{self._plugin.NAME} reached configured plugin_depth_limit",
                ),
                date_start=run_start,
                date_end=run_start,
                runtime=0,
            )
            logger.warning(f"{self._plugin.NAME}: hit max depth OPT_OUT for '{event.entity.sha256}'")
            yield (result, None)
            return  # Fetch next job

        job = models.Job(event=event)
        try:
            job.load_streams(dp=self._network.api, local=local)
        except (StorageError, FileNotFoundError):
            result = JobResult(
                state=State(
                    State.Label.ERROR_INPUT,
                    failure_name="Error opening data stream",
                    message=traceback.format_exc(),
                ),
                date_start=run_start,
                date_end=run_start,
                runtime=0,
            )
            logger.warning("Error opening data stream for %s" % event.entity.sha256)
            yield (result, None)
            return  # Fetch next job

        # run each multiplugin
        run_start = datetime.datetime.now(datetime.timezone.utc)

        # run first entrypoint to see if the whole plugin is opting out
        # first entrypoint == execute() function
        result = self._run_job_with_multiplugin(job, multiplugin=None, queue=queue)
        # start time should include multiprocessing code
        result.date_start = run_start
        if result.state.label not in azm.StatusEnumSuccess:
            # ending time should include multiprocessing code
            now = datetime.datetime.now(datetime.timezone.utc)
            result.date_end = now
            result.runtime = int((now - run_start).total_seconds())
            # yield heartbeats and other non-ok events
            yield result, None
            return  # fetch next job

        # Execute any multiplugins and always send the main result
        # NOTE - multiprocessing is not included in the time cost of these events
        try:
            for multiplugin in self._plugin.get_registered_multiplugins():
                if multiplugin is None:
                    continue
                yield self._run_job_with_multiplugin(job, multiplugin, queue), multiplugin
        finally:
            # the overall plugin finished without error, now need to raise event for it
            # send main plugin message with runtime covering all multiplugins and multiprocessing code
            now = datetime.datetime.now(datetime.timezone.utc)
            result.date_end = now
            result.runtime = int((now - run_start).total_seconds())
            yield result, None

    def _run_job_with_multiplugin(
        self, job: models.Job, multiplugin: Optional[str], queue: multiprocessing.Queue = None
    ) -> JobResult:
        """Run a multiplugin with the given event and streams."""
        try:
            # Don't enqueue job if multiplugin is None as it will have already been enqueued.
            if queue and multiplugin is not None:
                queue.put(
                    TaskModel(in_event=job.event, start_time_epoch=time.time(), multi_plugin_name=multiplugin),
                    timeout=QUEUE_PUT_TIMEOUT,
                )
            result = plugin_executor.run_plugin_with_job(self._plugin, job, multiplugin)
        except BaseException:
            logger.fatal("run_once terminated with uncaught exception")
            raise

        logger.info(
            "finish plugin=%s mp=%s state=%s file_format=%s size=%s sha256=%s"
            % (
                self._plugin.NAME,
                multiplugin,
                result.state.label.name,
                job.event.entity.file_format,
                job.event.entity.size,
                job.event.entity.sha256,
            )
        )
        return result



"""Contains a template test case class to simplify testing of Azul plugins.

Usage:
  from azul_runner import test_template
  from plugins_package.my_plugin import MyPlugin

  class TestMyPlugin(test_template.TestPlugin):
      PLUGIN_TO_TEST = MyPlugin

      def test_foo_case(self):
          result = self.do_execution(feats_in={...}, files_in=[('stream-name', '<filename>'), ...])
          self.assertJobResult(result, JobResult(...))
          ...
"""

import datetime
import importlib
import io
import os
import pathlib
import struct
import tempfile
import typing
import unittest
import warnings
from sys import gettrace
from typing import Any, ClassVar, Iterable, Optional, Type
from unittest import mock

import black
import cart
from azul_bedrock import dispatcher
from azul_bedrock import exceptions_bedrock as azbe
from azul_bedrock import models_network as azm
from azul_bedrock.exception_enums import ExceptionCodeEnum
from azul_bedrock.test_utils import file_manager

from azul_runner import coordinator, settings
from azul_runner.coordinator import Coordinator

from .. import local, monitor
from ..models import EventBase, Feature, FeatureValue, JobResult
from ..plugin import Plugin
from ..storage import DATA_HASH, StorageProxyFile

unittest.util._MAX_LENGTH = 2000
MANDATORY_CART_HEADER_LEN = struct.calcsize(cart.MANDATORY_HEADER_FMT)


class TestPlugin(unittest.TestCase):
    """Parent class to simplify testing of concrete plugins.

    Handles setting up and execution of plugin using a provided sample file/data bytes
    and input features.

    Usage:
        class TestFooPlugin(TestPlugin):
            PLUGIN_TO_TEST = FooPlugin

            def test_normal_execution(self):
                result = self.do_execution(
                    feats_in = {
                        'feature1': [val1, val2, val3, ...],
                        'feature2': [
                            FeatureValue('blah', tag='foo', ...),
                            ...
                        ]
                    },
                    file_in = [('content', 'path/to/file'), ... ],
                    data_in = [('stream_label', b'<bytes>'), ...],
                )
                ... Make assertions about results ...
    """

    PLUGIN_TO_TEST: ClassVar[Type[Plugin]]  # Override in subclass
    PLUGIN_TO_TEST_CONFIG: ClassVar[dict] = {}  # Override in subclass where necessary
    maxDiff = None

    def without_data(self, result: JobResult):
        """Deprecated. Return copy of result with cleared data content strings.

        Use self.assertJobResult(x, y) instead.
        """
        warnings.warn("deprecated call to without_data - this does nothing.", DeprecationWarning, stacklevel=2)
        return result

    def _strip_hash(self, result: JobResult):
        """Replace data blob hashes and child ids with numbers.

        Some libraries may produce inconsistent output data so this allows for unit tests
        to still succeed on an equality check.

        Crucially, this allows for checking equality of metadata the plugin generated,

        Usecase: In a test, a generated png file may contain a date resulting in a new
        hash every time the test is run.
        """
        result = result.model_copy()
        replaces = {}
        for i, k in list(enumerate(result.data)):
            replaces[k] = str(i)
            result.data[str(i)] = result.data.pop(k)

        def _nested(e: EventBase):
            """Replace data hash wherever it is found."""
            if e.parent:
                # parents have nested parents which need to be processed
                _nested(e.parent)
            if e.parent_sha256 in replaces:
                e.parent_sha256 = replaces[e.parent_sha256]
            if e.sha256 in replaces:
                if e.md5:
                    e.md5 = replaces[e.sha256]
                if e.sha512:
                    e.sha512 = replaces[e.sha256]
                if e.sha1:
                    e.sha1 = replaces[e.sha256]
                e.sha256 = replaces[e.sha256]

        for e in result.events:
            _nested(e)
            # parent entities don't have a data block
            for d in e.data:
                if d.hash in replaces:
                    d.hash = replaces[d.hash]

        return result

    @classmethod
    def setUpClass(cls) -> None:
        """Unittest method."""
        if cls is TestPlugin:
            raise unittest.SkipTest("Template class for plugin tests cannot be run directly")
        # Create a temporary directory for storing results that are transferred by monitor back to the parent process.
        cls.test_multi_process_temp_dir = tempfile.TemporaryDirectory(
            prefix="test_multi_process_temp_dir", ignore_cleanup_errors=True
        )
        cls._file_manager = file_manager.FileManager()

    @classmethod
    def tearDownClass(cls) -> None:
        """Teardown the multiprocess test directory."""
        # Cleanup the temporary directory used for transferring data from child process.
        cls.test_multi_process_temp_dir.cleanup()

        return super().tearDownClass()

    def setUp(self) -> None:
        """Unittest method."""
        # Tried to do this with just a setUpClass method, but for some reason unittest doesn't seem to run
        #  setUpClass when the class is imported to another file... but still tries to run the tests.
        if self.__class__ is TestPlugin:
            raise unittest.SkipTest("Template class for plugin tests cannot be run directly")

        self._results = []

    def tearDown(self) -> None:
        """Clean up resources."""
        # try to close handles
        for res in self._results:
            res.close()

    def load_test_file_bytes(self, sha256: str, description: str) -> bytes:
        """Download a file from a cache or virustotal and provide the raw bytes.

        Description is intended to give a description of what the file you are downloading is.
        This is added to the cache and useful for inline documentation about the file.
        """
        return self._file_manager.download_file_bytes(sha256)

    def load_test_file_path(self, sha256: str, description: str) -> pathlib.Path:
        """Download a file from a cache or virustotal and provide the path to a copy in temp.

        Description is intended to give a description of what the file you are downloading is.
        This is added to the cache and useful for inline documentation about the file.
        """
        return self._file_manager.download_file_path(sha256)

    def _get_location(self) -> str:
        """Return path to child class that implements this class."""
        # import child module
        module = type(self).__module__
        i = importlib.import_module(module)
        # get location to child module
        return i.__file__

    def _get_data(self, sub1: str, sub2: str = None) -> typing.BinaryIO:
        """Return expected path to test data."""
        # get folder where test is
        folder = os.path.split(self._get_location())[0]
        path = os.path.join(sub1, sub2) if sub2 else sub1
        path = os.path.join(folder, "data", path)
        return open(path, "rb")

    def load_local_raw(
        self,
        path: str,
        subpath: str = None,
        *,
        description: str,
    ) -> bytes:
        """Return contents of a local file that is known to be good from the data directory.

        Description is intended to give a description of what the file you are loading is.
        This is added to the cache and useful for inline documentation about the file.
        """
        with self._get_data(path, subpath) as istream:
            data = istream.read()
        return data

    def _unpack_cart_and_read(self, input_stream: typing.BinaryIO) -> bytes:
        """Conditionally unpacks CaRT files and reads bytes from provided filepath."""
        # Ensure read from start of the file
        input_stream.seek(0)
        header = input_stream.read(MANDATORY_CART_HEADER_LEN)
        # Reset the read head, incase it is not a CaRT file, the function can just return
        input_stream.seek(0)
        if cart.is_cart(header):
            unpacked = io.BytesIO()
            cart.unpack_stream(input_stream, unpacked)
            unpacked.seek(0)
            return unpacked.getvalue()
        else:
            return input_stream.read()

    def load_cart(self, path: str, subpath: str = None, *, description: str) -> bytes:
        """Unwrap and return a cart file from the data directory.

        Description is intended to give a description of what the file you are loading is.
        This is added to the cache and useful for inline documentation about the file.
        """
        with self._get_data(path, subpath) as istream:
            data = self._unpack_cart_and_read(istream)
        return data

    def assertJobResultsDict(
        self,
        actual: dict[str, JobResult],
        expected: dict[str, JobResult],
        *,
        inspect_data: bool = False,
        strip_hash: bool = False,
    ):
        """Compare two dictionaries of JobResult objects for compatibility.

        Will strip out data blobs as they are large and sha256 comparison should
        meet the needs of most testing. Deeper testing will require manual inspection
        of the blobs in the job result or call assertFormatted instead.

        If strip_hash is set, 'in1' data stream hashes will be replaced with a number.
        This is useful when libraries generate inconsistent output due to versioning
        if you don't care for exact matches.
        """
        actual, expected = self._baseAssertJobResult(
            actual, expected, inspect_data=inspect_data, strip_hash=strip_hash
        )
        self.assertReprEqual(actual, expected)

    def assertJobResult(
        self, actual: JobResult, expected: JobResult, *, inspect_data: bool = False, strip_hash: bool = False
    ):
        """Compare two JobResult objects for compatibility.

        Will strip out data blobs as they are large and sha256 comparison should
        meet the needs of most testing. Deeper testing will require manual inspection
        of the blobs in the job result or call assertFormatted instead.

        If strip_hash is set, 'in1' data stream hashes will be replaced with a number.
        This is useful when libraries generate inconsistent output due to versioning
        if you don't care for exact matches.
        """
        actual, expected = self._baseAssertJobResult(
            {None: actual}, {None: expected}, inspect_data=inspect_data, strip_hash=strip_hash
        )
        self.assertReprEqual(actual[None], expected[None])

    def _baseAssertJobResult(
        self,
        actual_dict: dict[str, JobResult],
        expected_dict: dict[str, JobResult],
        *,
        inspect_data: bool = False,
        strip_hash: bool = False,
    ) -> dict[str, JobResult, dict[str, JobResult]]:
        # must not modify data block of original JobResult
        for actual_key in actual_dict:
            actual_dict[actual_key] = actual_dict[actual_key].model_copy()
            # process bytesio, in either case do not close these handles
            data2 = {}
            if inspect_data:
                # convert data to bytes
                for k, v in actual_dict[actual_key].data.items():
                    v.seek(0, 2)
                    size = v.tell()
                    if size > 10_000:
                        raise Exception(f"result data {k=} is too large to inspect {size=}")
                    v.seek(0)
                    data2[k] = v.read()
                    # reset pointer in case another thing wants to read the file
                    v.seek(0)
            else:
                # strip data
                for k in actual_dict[actual_key].data:
                    data2[k] = b""
            actual_dict[actual_key].data = data2

            if strip_hash:
                actual_dict[actual_key] = self._strip_hash(actual_dict[actual_key])

        # Normalise both data sources for simpler comparison.
        # Converts all the expected outputs feature values to their encoded equivalent and sorts them again.
        # Needed as supplied data could be something like a raw datetime or integer and it needs to test consistently.
        sources = [expected_dict, actual_dict]
        for source in sources:
            for expected in source.values():
                for evnt in expected.events:
                    for feat_val_list in evnt.features.values():
                        for i, feat_val in enumerate(feat_val_list):
                            feat_val_list[i] = FeatureValue(
                                feat_val.value_encoded(),
                                label=feat_val.label,
                                offset=feat_val.offset,
                                size=feat_val.size,
                            )
                        # ensure consistent sorting
                        feat_val_list.sort(key=lambda x: x.value)

        return actual_dict, expected_dict

    def assertReprEqual(self, actual, expected):
        """Format input repr() as multiline strings so unittest diff works on objects.

        This requires the input repr() to be valid python code and is intended for use with
        model Run, Event, FeatureValue, etc.
        """
        try:
            self.assertEqual(actual, expected)
        except AssertionError:
            print(
                "The below printout can form the base of your test, "
                "but you must double check the output is as expected.\n"
            )
            print(repr(actual))
            print("")
            # use black to render the difference between the two
            # exploiting the fact that the repr() is valid python code
            string1 = black.format_str(repr(actual), mode=black.Mode())
            string2 = black.format_str(repr(expected), mode=black.Mode())
            self.assertEqual(string1, string2)

    def test_00_testplugin_set(self):
        """Assert that PLUGIN_TO_TEST is set to something."""
        self.assertTrue(hasattr(self, "PLUGIN_TO_TEST"))
        self.assertIsNotNone(self.PLUGIN_TO_TEST)

    def test_00_execute_present(self):
        """Check that the execute() method is present."""
        self.assertTrue(hasattr(self.PLUGIN_TO_TEST, "execute"))
        self.assertTrue(callable(self.PLUGIN_TO_TEST.execute))

    def test_00_registration_info(self):
        """Check that the required plugin metadata (registration info) is present."""
        found_features: set[str] = set()  # Record of feature names seen, to track duplicates

        def make_feature_assertions(f: Feature):
            self.assertIsInstance(f, Feature)
            self.assertIsInstance(f.name, str)
            self.assertNotEqual(f.name, "")
            self.assertNotIn(f.name, found_features, "Duplicate feature defined: %s" % f.name)
            found_features.add(f.name)
            self.assertIsInstance(f.desc, str)
            self.assertNotEqual(f.desc, "")

        for a in ("VERSION", "FEATURES"):
            self.assertTrue(hasattr(self.PLUGIN_TO_TEST, a), "Plugin requires %s attribute" % a)

        try:
            # Use object getattribute for this one to bypass the parent-class accumulation in Plugin's getattr
            output_feats = object.__getattribute__(self.PLUGIN_TO_TEST, "FEATURES")
        except AttributeError:
            # If the plugin doesn't define FEATURES, ensure that one of its parents does, then
            #  set output_feats to an empty list (the parents output features don't need to be validated.)
            self.assertTrue(
                self.PLUGIN_TO_TEST.FEATURES,
                "Neither %s not its parents defines FEATURES!" % self.PLUGIN_TO_TEST,
            )
            output_feats = []

        plugin = self.PLUGIN_TO_TEST(self.PLUGIN_TO_TEST_CONFIG)
        req_content = plugin.cfg.filter_data_types

        # SECURITY_LEVEL doesn't have to be present, but if it is it must be None or a string
        self.assertIsInstance(getattr(self.PLUGIN_TO_TEST, "SECURITY_LEVEL", None), (type(None), str, list, dict))

        self.assertIsInstance(self.PLUGIN_TO_TEST.VERSION, str)

        self.assertIsInstance(req_content, dict)
        for k, items in req_content.items():
            self.assertIsInstance(k, str)
            self.assertIsInstance(items, list)
            for i in items:
                self.assertIsInstance(i, str)

        self.assertIsInstance(output_feats, Iterable)

        for item in output_feats:
            make_feature_assertions(item)

    def do_execution_multi(self, *args, **kwargs) -> dict[str, JobResult]:
        """Typing helper."""
        return self.do_execution(*args, **kwargs)

    def do_execution(
        self,
        *,
        feats_in: Optional[list[azm.FeatureValue]] = None,
        files_in: Optional[list[tuple[str, str]]] = None,
        data_in: Optional[list[tuple[str, bytes]]] = None,
        ent_id: Optional[str] = "test_entity",
        config: Optional[dict[str, dict[str, Any]]] = None,
        keep_times: Optional[bool] = False,
        keep_feature_types: Optional[bool] = False,
        verify_input_content: Optional[bool] = True,
        plugin_class: Optional[Type[Plugin]] = None,
        entity_attrs: Optional[dict] = None,
        submission_settings: Optional[dict] = None,
        provided_coordinator: Optional[Coordinator] = None,
        no_multiprocessing: Optional[bool] = False,
        check_consistent_augmented_stream: bool = True,
    ) -> JobResult:
        """Calls the plugin's execute method and returns the results.

        If plugin is specified, that plugin is used instead.

        Uses either the given input features and either a (local) input file or a specified byte sequence.

        `files_in` and `data_in` may both be None if the plugin doesn't care about the input entity content.
        If present, they should contain tuples of ('stream_type', ['filename'|b'content']) respectively.

        If `ent_id` is specified, it will be passed to the plugin, otherwise a hash of the data is used.

        If `config` is specified, it is passed directly to the plugin init.

        If `keep_times` is True, the results include date_start, date_end and runtime; otherwise, these are removed.

        If `return_plugin_instance` is True, result['_plugin_instance'] will be set to the plugin that just ran,
        so the caller can make assertions about the plugin state. Default is to not return _plugin_instance.

        If `verify_input_content` is True (the default), an assertion will be raised if the input test content
        does not match the plugin's data restriction.

        `entity_attrs`: Optionally specify additional attrs for the entity model.

        `provided_coordinator`: provide a coordinator instance instead of the default monitor which tracks memory,
        and timeouts and runs in a subprocess. (doesn't apply config if provided!)
        This is typically used in mocking situations where you need more state out of the plugin then you can get from
        the default running method, because it's occurring in a subprocess.
        NOTE - use as a last resort!

        `no_multiprocessing`: bypass running as a multiprocess, this should be used sparingly and is useful
        when mocking restapi's with libraries like respx.

        `check_consistent_augmented_stream`: re-runs the plugin if streams (augmented or child) are produced.
        This is to ensure a stable interface but disabling this will half the run time for those tests.

        """
        if feats_in is None:
            feats_in = []
        if files_in is None:
            files_in = []
        if data_in is None:
            data_in = []
        if config is None:
            config = self.PLUGIN_TO_TEST_CONFIG
        if entity_attrs is None:
            entity_attrs = {}
        if plugin_class is None:
            plugin_class = self.PLUGIN_TO_TEST

        # Default to raising errors on comms failure instead of retrying.
        # This can be overridden by setting 'request_retry_count'.
        config.setdefault("request_retry_count", 0)

        # Default to not having a timeout when debugging to prevent interruption.
        # This can be overridden by setting 'run_timeout'.
        if gettrace() is not None:
            config.setdefault("run_timeout", 0)

        input_data = []
        if (not ent_id or ent_id == "test_entity") and data_in:
            # use data hash if content is present.
            for k, v in data_in:
                if k == azm.DataLabel.CONTENT:
                    ent_id = DATA_HASH(v).hexdigest()
                    input_data = [local.gen_api_content(io.BytesIO(v))]
                    break

        # generate the event
        entity = azm.BinaryEvent.Entity(sha256=ent_id, datastreams=input_data, features=feats_in, **entity_attrs)

        # fill in properties from data
        if input_data:
            d = input_data[0]
            entity.datastreams = input_data
            entity.size = input_data[0].size
            entity.sha1 = d.sha1
            entity.sha256 = d.sha256
            entity.sha512 = d.sha512
            entity.md5 = d.md5
            entity.file_format = d.file_format

        event = local.gen_event(entity=entity)
        if submission_settings:
            event.source.settings = submission_settings

        # Set up data streams
        # Constraint: input files are read into memory, so you cannot test with massive sample files.
        datastreams: list[StorageProxyFile] = []

        for label, fname in files_in:
            with open(fname, "rb") as f:
                data = f.read()
            datastreams.append(
                StorageProxyFile(
                    source="local",
                    label=label,
                    hash=fname,
                    init_data=data,
                    file_info=local.gen_api_content(io.BytesIO(data), label=label),
                )
            )

        for label, data in data_in:
            datastreams.append(
                StorageProxyFile(
                    source="local",
                    label=label,
                    hash="test-data",
                    init_data=data,
                    file_info=local.gen_api_content(io.BytesIO(data), label=label),
                )
            )
        try:
            # if no_timeout_or_heartbeat_coordinator:
            if provided_coordinator:
                monitor_m_or_coord = provided_coordinator
            elif no_multiprocessing:
                monitor_m_or_coord = coordinator.Coordinator(plugin_class, settings.parse_config(plugin_class, config))
            else:
                monitor_m_or_coord = monitor.Monitor(plugin_class, config)
            # Verify that the specified data actually matches the provided test input
            if verify_input_content:
                local.validate_streams(datastreams, monitor_m_or_coord._plugin.cfg)

            try:
                test_dir_name = self.test_multi_process_temp_dir.name
            except AttributeError as e:
                raise AttributeError(
                    "Have you overridden setUpClass or tearDownClass without calling 'super().setUpClass()'."
                    + f" Missing test_multi_process_temp_dir with error - {e}"
                ) from e

            if isinstance(monitor_m_or_coord, Coordinator):
                results = monitor_m_or_coord.run_once(event, datastreams)
            else:
                results = monitor_m_or_coord.run_once(event, datastreams, test_dir_name)

            # generate status event that would be sent over network
            # during plugin execution errors at this step cause the plugin to crash
            for multiplugin, run in results.items():
                monitor_m_or_coord._network.api.submit_binary = lambda x, y, z: local.gen_api_content(z, label=y)
                mm = monitor_m_or_coord._network.api.submit_events = mock.MagicMock()
                monitor_m_or_coord._network.ack_job(event, run, multiplugin)
                # append result to tearDown list
                self._results.append(run)
                status = mm.call_args[0][0][0]
                if (
                    eventlen := len(status.model_dump_json(exclude_defaults=True).encode())
                ) > dispatcher.MAX_MESSAGE_SIZE:
                    raise azbe.NetworkDataException(
                        ref=f"event produced by plugin was too large: {eventlen}b"
                        + f" > {dispatcher.MAX_MESSAGE_SIZE}b",
                        internal=ExceptionCodeEnum.TestRunnerExecutionEventTooLarge,
                        parameters={"length_of_event": eventlen, "max_message_size": dispatcher.MAX_MESSAGE_SIZE},
                    )

                self.assertTrue(status)

            # Checks that the time-sensitive values of results are within tolerance, then discards them.
            # End date of overall plugin should be within a few seconds of now (upped from 3 to 5 seconds for timeouts)
            self.assertLess(
                (datetime.datetime.now(datetime.timezone.utc) - results[None].date_end).total_seconds(), 10
            )
            for res in results.values():
                # Runtime should be within 1s due to int rounding
                self.assertAlmostEqual(res.runtime, (res.date_end - res.date_start).total_seconds(), delta=1)
                if not keep_times:
                    res.runtime = None
                    res.date_start = None
                    res.date_end = None
                if not keep_feature_types:
                    res.feature_types = []

            # Verify that the plugin gives consistent and child/augmented streams on repetitive runs.
            if check_consistent_augmented_stream:
                does_result_have_data = any([len(r.data) > 0 for r in results.values()])
                if does_result_have_data:
                    # reset all input streams
                    for stream in datastreams:
                        stream.seek(0)
                    if isinstance(monitor_m_or_coord, Coordinator):
                        results2 = monitor_m_or_coord.run_once(event, datastreams)
                    else:
                        results2 = monitor_m_or_coord.run_once(event, datastreams, test_dir_name)

                    self.assertIsNotNone(
                        results2, "Result 2 should not be none as there is augmented or child streams present."
                    )
                    # Verify all the streams for each multiplugin are equal.
                    for m_plugin, result in results.items():
                        self.assertCountEqual(
                            list(results2[m_plugin].data.keys()),
                            list(result.data.keys()),
                            "Inconsistent data streams, your plugin should return the same results when it's re-run.",
                        )
        finally:
            for ds in datastreams:
                if not ds.closed:
                    ds.close()

        # handle simple plugin results (backwards compatibility)
        if len(results) == 1:
            # return single results dict
            return list(results.values())[0]
        else:
            # return dictionary of multiplugin results dicts
            return results



"""Init for all tests."""

import logging
import multiprocessing

# Preload azul_runner, this dramatically improves test speed when running in forkserver.
multiprocessing.set_forkserver_preload(["azul_runner"])

# Ensures forkserver which is used which is how all future code is expected to run (python 3.14+)
try:
    multiprocessing.set_start_method("forkserver")
except Exception:
    pass
# Ensure logging is enabled during tests.
logging.basicConfig(force=True)



"""Data structures used by plugins to record job results, such as features, info, data."""

from __future__ import annotations

import datetime
import io
import logging
import tempfile
import typing
from enum import Enum
from functools import total_ordering
from math import inf
from typing import Annotated, Any, ClassVar, Type

from azul_bedrock import dispatcher
from azul_bedrock import models_network as azm
from pydantic import (
    ConfigDict,
    PlainSerializer,
    PrivateAttr,
    SerializerFunctionWrapHandler,
    WrapSerializer,
    field_validator,
)

from . import storage

logger = logging.getLogger(__name__)


class BaseModelStrict(azm.BaseModelStrict):
    """Alters bedrock base model to have no default import namespace."""

    _DEFAULT_IMPORT = None


class ModelError(TypeError):
    """Error when validating that data matches the model."""

    pass


class Filepath(str):
    """A string value that will be parsed by Elastic as a file path."""

    pass


class Uri(str):
    """A string value that will be parsed by Elastic as a URI."""

    pass


# As inheritance for enums is impossible, this is used to override repr() and str()
# to render a valid reference to a state, for ease-of-use of the testing harness.
# WARNING - This directly alters the modules azm.StatusEventEnum so side effects will occur if you
# use that directly.
_status_event_enum_override = azm.StatusEnum
_status_event_enum_override.__repr__ = lambda self: f"State.Label.{self.name}"
_status_event_enum_override.__str__ = lambda self: f"State.Label.{self.name}"


class State(BaseModelStrict):
    """Data class recording the result of a plugin run."""

    Label: ClassVar[Type[azm.StatusEnum]] = _status_event_enum_override
    label: "State.Label"  # COMPLETED, or category of reason for non-execution
    failure_name: str | None = None  # Brief title of problem, eg 'OSError' or 'opt-out'
    message: str | None = None  # Any other output messages (eg traceback for exception)

    def __init__(self, label: Label = Label.COMPLETED, failure_name: str = None, message: str = None):
        if label == State.Label.OPT_OUT:
            if not message:
                message = "No opt-out reason was provided."
        super().__init__(**dict(label=label, failure_name=failure_name, message=message))

    def __repr__(self):
        """Custom pydantic repr."""
        return azm.repr_reproduce(self, required=["label"])

    __str__ = __repr__


State.model_rebuild()

# Lookup table for converting types to type functions.
feat_type_lookup_table = {
    "integer": int,
    "float": float,
    "string": str,
    "binary": bytes,
    "datetime": datetime.datetime,
    "filepath": str,
    "uri": str,
}


@total_ordering
class Feature(BaseModelStrict):
    """Data class to hold the definition of an output feature for a plugin."""

    name: str
    desc: str
    type: str
    # accepted class for values of the feature
    _typeref: type = PrivateAttr()

    def __init__(self, name, desc, type: azm.FeatureType = azm.FeatureType.String) -> None:
        # FUTURE rename parameter 'type' to 'kind' due to shadowing builtin 'type' function
        kind = type
        if kind not in azm.FeatureType:
            orig_kind = kind
            logger.warning(f"feature {name} should use a FeatureType.<option> enum entry for type, not {orig_kind}")
            # convert legacy types to enum
            try:
                kind = {
                    int: azm.FeatureType.Integer,
                    float: azm.FeatureType.Float,
                    str: azm.FeatureType.String,
                    bytes: azm.FeatureType.Binary,
                    datetime.datetime: azm.FeatureType.Datetime,
                    Filepath: azm.FeatureType.Filepath,
                    Uri: azm.FeatureType.Uri,
                }[kind]
            except KeyError:
                raise ValueError(
                    f"feature {name} should use a FeatureType.<option> enum entry for type, not {orig_kind}"
                ) from None

        super().__init__(**dict(name=name, desc=desc, type=kind))

        # Has to be after super()
        self._typeref = feat_type_lookup_table[kind]

    def __hash__(self):
        """Hash function."""
        return hash(self.name)

    def __eq__(self, other: Feature) -> bool:
        """Equal function."""
        return NotImplemented if not isinstance(other, Feature) else (self.name == other.name)

    def __lt__(self, other: Feature) -> bool:
        """Less than function."""
        return NotImplemented if not isinstance(other, Feature) else (self.name < other.name)

    @property
    def typeref(self):
        """Return a class reference to the features type."""
        return self._typeref


@total_ordering
class FeatureValue(BaseModelStrict):
    """Data class to store the value of a feature, possibly with a label, offset, and/or size."""

    value: Annotated[azm.VALUE_DECODED, PlainSerializer(lambda v: azm.value_encode(v), return_type=azm.VALUE_DECODED)]
    label: str | None = None
    offset: int | None = None
    size: int | None = None

    def __init__(self, value=None, **kwargs):
        # convert legacy types to strings
        if isinstance(value, (Filepath, Uri)):
            value = str(value)
            logging.warning("Filepath(x) and Uri(x) are deprecated, use string instead.")
        kwargs["value"] = value
        super().__init__(**kwargs)

    def __repr__(self):
        """Custom pydantic repr."""
        # Print 'FV' instead of 'FeatureValue' because it is so common and takes up too much space otherwise.
        return azm.repr_reproduce(self, "FV", ["value"])

    __str__ = __repr__

    @property
    def _comparison_tuple(self):
        return (
            str(type(self.value)),
            self.value,
            "" if self.label is None else self.label,
            -inf if self.offset is None else self.offset,  # 'None' sorts before all values
            -inf if self.size is None else self.size,
        )

    def __eq__(self, other: FeatureValue) -> bool:
        """Explicitly implement equality only for FeatureValue to FeatureValue instances."""
        return (
            NotImplemented
            if not isinstance(other, FeatureValue)
            else (self._comparison_tuple == other._comparison_tuple)
        )

    def __lt__(self, other: FeatureValue) -> bool:
        """Explicitly implement less than for FeatureValue to FeatureValue instances."""
        return (
            NotImplemented
            if not isinstance(other, FeatureValue)
            else (self._comparison_tuple < other._comparison_tuple)
        )

    def value_encoded(self) -> str:
        """Return encoded value."""
        return azm.value_encode(self.value)

    model_config = ConfigDict(frozen=True)


# shortcut to commonly used class
FV = FeatureValue


@total_ordering
class EventData(BaseModelStrict):
    """Binary data and metadata."""

    # Ensure enum values are made into strings
    model_config = ConfigDict(use_enum_values=True)

    hash: str
    label: azm.DataLabel
    language: str | None = None

    def __eq__(self, other: EventData) -> bool:
        """Explicitly implement equality only for EventData to EventData instances."""
        return NotImplemented if not isinstance(other, EventData) else (self.hash == other.hash)

    def __lt__(self, other: EventData) -> bool:
        """Explicitly implement less than for EventData to EventData instances."""
        return NotImplemented if not isinstance(other, EventData) else (self.hash < other.hash)


class EventBase(azm.FileInfo):
    """Simple base for partial event."""

    _DEFAULT_IMPORT = None

    parent: EventParent | None = None
    parent_sha256: str | None = None
    sha256: str
    relationship: dict = {}

    def __init__(self, *args, **kwargs):
        # Backwards compatibility mapping for old plugin unittests to avoid having to update all the old tests.
        # (Maps the old name of fields to the new name).
        renames = [
            ("entity_id", "sha256"),
            ("parent_hash", "parent_sha256"),
            ("entity_type", "model"),
            ("file_type_al", "file_format"),
            ("mime_type", "mime"),
            ("mime_magic", "magic"),
        ]
        did = []
        for old, new in renames:
            if old in kwargs:
                kwargs[new] = kwargs.pop(old)
                did.append((old, new))

        # model is implicitly binary so drop this
        kwargs.pop("model", None)
        if did:
            logging.warning(f"Test Event() creation should replace args: {did}")

        super().__init__(*args, **kwargs)


class EventParent(EventBase):
    """Minified event preventing the need for events to track parent events."""

    filename: str | None = None
    language: str | None = None


class Event(EventBase):
    """Details of a partial event generated by a plugin.

    Note that a lot of properties may be unset, the original job event should be used to fill in these details.
    """

    data: list[EventData] = []
    features: dict[str, list[FeatureValue]] = {}
    info: dict = {}

    # track binary data in shared dictionary
    _shared_data: dict[str, typing.BinaryIO] = PrivateAttr()
    # track events in shared list
    _events: list[Event] = PrivateAttr()
    # track children of the current event so duplicate children are not added
    _children: dict[tuple[str, str, str], Event] = PrivateAttr({})

    @field_validator("features")
    @classmethod
    def _sort_features(cls, f):
        """Sort features on creation."""
        # sort feature values
        for v in f.values():
            v.sort()

        # sort feature names
        return dict(sorted(f.items()))

    def sort(self):
        """Sort contents of event."""
        # sort feature values
        for k in self.features.keys():
            self.features[k].sort()

        # sort feature names
        self.features = dict(sorted(self.features.items()))

    def set_stores(self, binaries: dict, events: list):
        """Update stores shared between all events."""
        self._shared_data = binaries
        self._events = events

    def as_parent(self) -> EventParent:
        """Generate an EventParent instance."""
        filenames = self.features.get("filename", [])
        # must sort before getting value, for consistency
        filename = sorted(filenames)[0].value if filenames else None
        lang = [x.language for x in self.data if x.label == "content" and x.language]
        # must sort before getting value, for consistency
        lang = sorted(lang)[0] if lang else None
        raw = self.model_dump()
        for k in ["data", "features", "info"]:
            # drop extra properties
            raw.pop(k, None)
        return EventParent(
            **raw,
            filename=filename,
            language=lang,
        )

    def keep(self) -> bool:
        """True if event is worth emitting from plugin."""
        # keep all child events
        # keep top level events if features, data or info are set
        return self.parent or (self.features or self.data or self.info)

    def add_data(self, label: azm.DataLabel, tags: dict[str, str], data: bytes) -> str:
        """Add data to the event and return data hash.

        Results in memory issues if data is large.
        """
        if len(data) > 2**24:
            logger.warning(f"adding large file as bytestring, which is bad for memory usage {label=}")
        return self.add_data_file(label, tags, io.BytesIO(data))

    def add_data_file(self, label: azm.DataLabel, tags: dict[str, str], data_file: typing.BinaryIO) -> str:
        """Add data file to the event and return data hash.

        File handle may be closed afterward.
        """
        # copy data file to temp file so that caller may close file handle and/or cleanup
        # FUTURE send to dispatcher straight away as this wastes disk space and time
        #   would require a harness for local execution and testing
        wrote_bytes = 0
        tmpfile = tempfile.TemporaryFile("r+b")
        data_file.seek(0)
        while True:
            # 100kb at a time
            buf = data_file.read(100_000)
            if not buf:
                break
            tmpfile.write(buf)
            wrote_bytes += len(buf)
        tmpfile.seek(0)

        if not wrote_bytes:
            raise ValueError("tried to add data file with 0 bytes")

        data_hash = storage.calc_stream_hash(data_file)
        if data_hash in [x.hash for x in self.data]:
            return
        data_file.seek(0)
        self.data.append(EventData(hash=data_hash, label=label, language=tags.pop("language", None)))
        if tags:
            raise ValueError(f"unknown tags for data {tags}")
        self._shared_data[data_hash] = tmpfile
        return data_hash

    def add_text(self, text: str, language: str = None) -> str:
        """Add a text stream to the current binary.

        Used for a report or other textual artifact related to the current plugin run. The text will be viewable
         in the UI.

        Optionally, you can specify the 'language' property to facilitate syntax highlighting in the Azul UI.
         This should be a language name supported by prism.js - eg, 'html', 'js'/'javascript', 'bash'/'shell',
         'c', 'dotnet', 'php', 'go', 'powershell', 'python', 'regex', 'vb'/'vba'
        """
        return self.add_data(azm.DataLabel.TEXT, {"language": language} if language else {}, text.encode("utf-8"))

    def add_feature_values(self, key: str, values: Any):
        """Add feature values to plugin result."""
        if not isinstance(values, (list, set)):
            values = [values]
        elif len(values) == 0:
            # Ignore empty lists and sets for feature values.
            logger.warning(f"ignoring empty list for feature {key}")
            return

        # normalise types
        values_fixed = []
        for value in values:
            if value is None:
                # ignore None values, likely the result of something like {}.get("myvalue"), which is common pattern
                logger.warning(f"ignoring None value for feature {key}")
                continue
            if not isinstance(value, FeatureValue):
                value = FeatureValue(value)
            values_fixed.append(value)
        self.features.setdefault(key, []).extend(values_fixed)

    def add_many_feature_values(self, features: dict[str, Any]):
        """Add a dictionary of feature values to the main event."""
        for k, v in features.items():
            self.add_feature_values(k, v)

    def add_info(self, info: dict):
        """Add info to event, overwriting previous info."""
        self.info = info

    def _add_child(self, sha256: str, relationship: dict, *, parent_sha256: str = None) -> Event:
        """Add child event to current event."""
        uniq = (parent_sha256, sha256)
        if uniq not in self._children:
            ret = Event(
                parent=self.as_parent(),
                parent_sha256=parent_sha256,
                sha256=sha256,
                relationship=relationship,
            )
            ret.set_stores(self._shared_data, self._events)
            self._events.append(ret)
            self._children[uniq] = ret
        return self._children[uniq]

    def add_child_with_data(self, relationship: dict, data: bytes, *, label: azm.DataLabel = azm.DataLabel.CONTENT):
        """Add child of current event using data hash as sha256.

        Results in memory issues if data is large.
        """
        return self.add_child_with_data_file(relationship, io.BytesIO(data), label=label)

    def add_child_with_data_file(
        self, relationship: dict, data_file: typing.BinaryIO, *, label: azm.DataLabel = azm.DataLabel.CONTENT
    ):
        """Add child of current event using data hash as sha256."""
        # data hash is child id
        data_hash = storage.calc_stream_hash(data_file)
        c = self._add_child(data_hash, relationship)
        # note - calculates hash again (inefficient)
        c.add_data_file(label, {}, data_file)
        return c


def custom_binary_serialize(data: dict[str, typing.BinaryIO | bytes], orig_serializer: SerializerFunctionWrapHandler):
    """Reads binary data before serialization to prevent data loss."""
    result = dict()
    for row in data:
        if issubclass(type(data[row]), io.IOBase):
            # FUTURE will overflow memory if large
            result[row] = data[row].read()
            data[row].seek(0)
        else:
            result[row] = data[row]
    return orig_serializer(result)


class JobResult(BaseModelStrict):
    """Contains results of a run of a plugin."""

    state: State
    events: list[Event] = []
    # data values are only 'bytes' during testing
    data: Annotated[dict[str, typing.BinaryIO | bytes], WrapSerializer(custom_binary_serialize)] = {}
    feature_types: list[Feature] = []
    runtime: int | None = None
    date_start: datetime.datetime | None = None
    date_end: datetime.datetime | None = None

    @property
    def main(self) -> Event:
        """Return main event."""
        return self.events[0] if len(self.events) else None

    def close(self):
        """Close any JobResult data handles."""
        for d in self.data:
            if issubclass(type(d), io.IOBase) and not self.data[d].closed:
                self.data[d].close()

    model_config = ConfigDict(arbitrary_types_allowed=True)


class Job(BaseModelStrict):
    """Contains event and data streams for plugin to process."""

    # event for the job
    event: azm.BinaryEvent

    @property
    def id(self) -> str:
        """Entity id."""
        return self.event.entity.sha256

    # Input content streams as file-like objects
    _streams: list[storage.StorageProxyFile] = PrivateAttr(None)

    # helpers
    def load_streams(self, *, dp: dispatcher.DispatcherAPI = None, local: list[storage.StorageProxyFile] = None):
        """Load streams from network or local files."""
        self._streams = []
        if local:
            self._streams = local
        elif dp:
            # initialise data streams
            if not self.event.entity.datastreams:
                return
            for stream in self.event.entity.datastreams:
                self._streams.append(
                    storage.StorageProxyFile(
                        source=self.event.source.name,
                        label=stream.label,
                        hash=stream.sha256,
                        dp=dp,
                        allow_unbounded_read=True,
                        file_info=stream,
                    )
                )

    def get_data(self, label: azm.DataLabel = azm.DataLabel.CONTENT) -> storage.StorageProxyFile:
        """Return and ensure only one stream with label."""
        streams = self.get_all_data(label)
        if len(streams) > 1:
            raise ValueError(f'more than one "{label}" stream for entity {self.id}')
        return streams[0] if streams else None

    def get_all_data(self, label: azm.DataLabel = None, file_format: str = None) -> list[storage.StorageProxyFile]:
        """Return data streams, optionally filtered by label and/or file_format."""
        if not self._streams:
            return []
        streams = self._streams
        if label:
            streams = [ds for ds in streams if ds.file_info.label == label]
        if file_format:
            streams = [ds for ds in streams if ds.file_info.file_format == file_format]
        return streams


class TaskExitCodeEnum(Enum):
    """Exit code enum."""

    COMPLETED = 0  # Process has completed nicely (probably job_limit reached).
    TERMINATE = 10  # An error has occurred that isn't fixing and the process should fully exit.
    RECREATE_PLUGIN = 20  # Process needs to be re-created (probably a git change detected).


class TaskModel(BaseModelStrict):
    """Model that holds the current BinaryEvent being processed by a plugin."""

    in_event: azm.BinaryEvent
    # Time since start of epoch (e.g output of time.time())
    start_time_epoch: float
    multi_plugin_name: str | None = None



"""Test cases that verify monitor's temp deletion works as expected."""

import os
import tempfile
import unittest

from azul_bedrock import models_network as azm

from azul_runner import local, monitor
from azul_runner.models import Job

from . import plugin_support as sup


class DeleteTempDirectoryTests(unittest.TestCase):
    def test_delete_tempdirectory_contents(self):
        file1 = tempfile.NamedTemporaryFile(delete=False, delete_on_close=False)
        file1.write(b"abc")
        file1.close()
        file_path = file1.name
        self.assertTrue(os.path.exists(file_path))
        monitor.Monitor.delete_tempfiles()
        self.assertFalse(os.path.exists(file_path))

    def base_delete_file_with_custom_prefix(self, file_name: str, prefix: str):
        # Create a file manually and confirm it's deleted
        file_to_create = os.path.join(tempfile.gettempdir(), file_name)
        with open(file_to_create, "w") as f:
            f.write("ABCDEFGHIJKLMNOPQRSTUVWXYZ")

        self.assertTrue(os.path.exists(file_to_create))
        # Default delete doesn't find the file
        monitor.Monitor.delete_tempfiles()
        self.assertTrue(os.path.exists(file_to_create))
        # Providing prefix with different case still finds the file.
        monitor.Monitor.delete_tempfiles(prefix)
        self.assertFalse(os.path.exists(file_to_create))

    def test_delete_prefix_lowercase_prefix_uppercase_filename(self):
        self.base_delete_file_with_custom_prefix("customTempFile", "customtemp")

    def test_delete_prefix_uppercase_prefix_lowercase_filename(self):
        self.base_delete_file_with_custom_prefix("customtempfile", "customTemp")

    def test_delete_same_case_on_prefix_and_filename(self):
        self.base_delete_file_with_custom_prefix("customTempFile", "customTempFile")

    class DPTemplatefileGetsCleared(sup.DummyPlugin):
        def execute(self, job: Job):
            with open(job.event.entity.info.get("file"), "wb") as f:
                f.write(b"abjksdlfjalksdf aslkdfjlkasdj lasdjf laksdfjlkasdf")

    def test_tempfile_gets_cleared(self):
        """Tests that bad temp files created by a plugin get cleared on re-create."""
        loop = monitor.Monitor(self.DPTemplatefileGetsCleared, {})
        first_dummy_file = os.path.join(tempfile.gettempdir(), "tmpdummyfile1")
        second_dummy_file = os.path.join(tempfile.gettempdir(), "tmpdummyfile2")
        entity = azm.BinaryEvent.Entity(sha256="id", datastreams=[], features=[], info={"file": first_dummy_file})
        loop.run_once(local.gen_event(entity))

        self.assertTrue(os.path.exists(first_dummy_file))
        entity = azm.BinaryEvent.Entity(sha256="id", datastreams=[], features=[], info={"file": second_dummy_file})
        loop.run_once(local.gen_event(entity))

        # First dummy file should have been deleted on next run when plugin was re-created.
        self.assertFalse(os.path.exists(first_dummy_file))

        self.assertTrue(os.path.exists(second_dummy_file))
        monitor.Monitor.delete_tempfiles()
        self.assertFalse(os.path.exists(second_dummy_file))



from __future__ import annotations

from azul_bedrock import models_network as azm

from azul_runner import (
    DATA_HASH,
    FV,
    BinaryPlugin,
    Event,
    EventData,
    EventParent,
    Feature,
    FeatureType,
    FeatureValue,
    Filepath,
    Job,
    JobResult,
    State,
    TestPlugin,
    Uri,
    add_settings,
)


class BinaryTestPlugin(BinaryPlugin):
    VERSION = "1.0"
    FEATURES = [
        Feature("sample_feature", "An output feature for the test plugin", type=FeatureType.String),
    ]

    def execute(self, job: Job) -> dict:
        pass


class TestBinaryTemplateDynamic(TestPlugin):
    PLUGIN_TO_TEST = BinaryTestPlugin

    class DPSimpleRun(BinaryTestPlugin):
        def execute(self, job):
            self.add_feature_values("sample_feature", ["completed ok"])

    def test_simple_run(self):
        """Check that the test plugin runs and returns a basic feature."""

        result = self.do_execution(
            data_in=[("content", b"dummy data")], plugin_class=TestBinaryTemplateDynamic.DPSimpleRun
        )
        self.assertEqual(result.state, State())
        self.assertEqual(result.main.features, {"sample_feature": [FeatureValue("completed ok")]})

    class DPLegacyStuff(BinaryTestPlugin):
        FEATURES = [
            Feature("a_filepath", "", type=Filepath),
            Feature("a_uri", "", type=Uri),
        ]

        def execute(self, job):
            self.add_feature_values("a_filepath", [FV(Filepath("/rootf/"))])
            self.add_feature_values("a_uri", [FV(Uri("http://abc.com"))])

    def test_legacy_stuff(self):
        """Check that legacy stuff works."""
        # legacy feature typing
        # legacy Filepath and Uri usage in plugin and tests
        result = self.do_execution(
            data_in=[("content", b"dummy data")],
            plugin_class=TestBinaryTemplateDynamic.DPLegacyStuff,
        )
        self.assertEqual(result.state, State())
        self.assertEqual(
            result.main.features,
            {
                "a_filepath": [FV(Uri("/rootf/"))],
                "a_uri": [FV(Uri("http://abc.com"))],
            },
        )

    class DPAddText(BinaryTestPlugin):
        def execute(self, job):
            self.add_text('print("Hello, world!")', "python")

    def test_add_text(self):
        """Check that binary.add_text functions correctly."""
        result = self.do_execution(
            data_in=[("content", b"dummy data")], plugin_class=TestBinaryTemplateDynamic.DPAddText
        )
        stream_id = DATA_HASH(b'print("Hello, world!")').hexdigest()
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="797bb0abff798d7200af7685dca7901edffc52bf26500d5bd97282658ee24152",
                        data=[
                            EventData(
                                hash=stream_id,
                                label=azm.DataLabel.TEXT,
                                language="python",
                            )
                        ],
                    )
                ],
                data={stream_id: b'print("Hello, world!")'},
            ),
            inspect_data=True,
        )

    class DPAddGrandchildInvalidData(BinaryTestPlugin):
        def execute(self, job):
            data = job.get_data()
            c = self.add_child_with_data({"r": "v"}, b"1")
            gc = c.add_child_with_data({"r": "v2"}, 45)

    class DPAddChildInvalidData(BinaryTestPlugin):
        def execute(self, job):
            self.add_child_with_data({"r": "v"}, 45)

    def test_add_child_invalid_data(self):
        """Check that add_binary_grandchild correctly tests for the child/parent to be present."""

        result = self.do_execution(
            data_in=[("content", b"dummy data")], plugin_class=TestBinaryTemplateDynamic.DPAddChildInvalidData
        )
        self.assertEqual(result.state, State(State.Label.ERROR_EXCEPTION, "TypeError", result.state.message))
        self.assertIn("a bytes-like object is required", result.state.message)

    def test_add_grandchild_invalid_data(self):
        """Check that add_binary_grandchild correctly tests for the child/parent to be present."""

        result = self.do_execution(
            data_in=[("content", b"dummy data")], plugin_class=TestBinaryTemplateDynamic.DPAddGrandchildInvalidData
        )
        self.assertEqual(result.state, State(State.Label.ERROR_EXCEPTION, "TypeError", result.state.message))
        self.assertIn("a bytes-like object is required", result.state.message)

    _child_data = b"foo"
    _child_id = DATA_HASH(_child_data).hexdigest()
    _gc_data = b"bar"
    _gc_id = DATA_HASH(_gc_data).hexdigest()

    class DPAddGrandchild(BinaryTestPlugin):
        def execute(self, job):
            data = job.get_data()
            c = self.add_child_with_data({"r": "v"}, TestBinaryTemplateDynamic._child_data)
            gc = c.add_child_with_data({"r": "v2"}, TestBinaryTemplateDynamic._gc_data)
            gc.add_feature_values("sample_feature", ["completed ok"])

    def test_add_grandchild(self):
        """Check that add_binary_grandchild correctly tests for the child/parent to be present."""
        result = self.do_execution(
            data_in=[("content", b"dummy data")], plugin_class=TestBinaryTemplateDynamic.DPAddGrandchild
        )
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        parent=EventParent(
                            sha256="797bb0abff798d7200af7685dca7901edffc52bf26500d5bd97282658ee24152",
                        ),
                        sha256=self._child_id,
                        relationship={"r": "v"},
                        data=[
                            EventData(
                                hash=self._child_id,
                                label=azm.DataLabel.CONTENT,
                            )
                        ],
                    ),
                    Event(
                        parent=EventParent(
                            parent=EventParent(
                                sha256="797bb0abff798d7200af7685dca7901edffc52bf26500d5bd97282658ee24152",
                            ),
                            sha256=self._child_id,
                            relationship={"r": "v"},
                        ),
                        sha256=self._gc_id,
                        relationship={"r": "v2"},
                        data=[
                            EventData(
                                hash=self._gc_id,
                                label=azm.DataLabel.CONTENT,
                            )
                        ],
                        features={"sample_feature": [FV("completed ok")]},
                    ),
                ],
                data={
                    "2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae": b"foo",
                    "fcde2b2edba56bf408601fb721fe9b5c338d10ee429ea04fae5511b68fbf8fb9": b"bar",
                },
            ),
            inspect_data=True,
        )



from __future__ import annotations

import datetime
import json
import logging
import multiprocessing
import time
import unittest
from typing import ClassVar

import httpx
import pytest
from azul_bedrock import models_network as azm

from azul_runner import State, monitor
from azul_runner.settings import add_settings

from . import mock_dispatcher as md
from . import plugin_support as sup


def dump(x):
    return json.loads(x.model_dump_json(exclude_defaults=True))


class DummySleepPlugin(sup.DummyPlugin):
    """Dummy plugin class that sleeps for 2 seconds before returning, to test timeouts with."""

    SETTINGS = add_settings(delay=(float, 2), delay_after_exception=(int, 0))

    @staticmethod
    def interruptable_sleep(secs):
        """Sleeps in bursts of 250ms so that exceptions can still interrupt us."""
        start = datetime.datetime.now()
        tick = 0.25
        while True:
            time.sleep(tick)
            elapsed = (datetime.datetime.now() - start).total_seconds()
            if elapsed >= secs:
                return
            elif secs - elapsed < tick:
                tick = secs - elapsed

    def execute(self, entity):
        if self.cfg.delay_after_exception:
            try:
                self.interruptable_sleep(self.cfg.delay)
            except monitor.PluginTimeoutError:
                self.interruptable_sleep(self.cfg.delay_after_exception)
        else:
            self.interruptable_sleep(self.cfg.delay)
        return


class TestPluginTimeouts(unittest.TestCase):
    """
    Tests the run_timeout functionality of Plugin.run_loop(), using the mock server.
    """

    mock_server: ClassVar[md.MockDispatcher]
    server: ClassVar[str]  # Endpoint to the mock server, suitable for passing to a plugin's config['server']
    dummy_log_handler: sup.DummyLogHandler  # Set for each instance in setUp()
    maxDiff = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.mock_server, cls.server = sup.setup_mock_dispatcher()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.mock_server.stop()
        cls.mock_server.kill()

    def setUp(self):
        lr = logging.getLogger("azul_runner")
        lr.setLevel(logging.WARNING)
        self.dummy_log_handler = sup.DummyLogHandler()
        lr.addHandler(self.dummy_log_handler)

    def _filter_error_from_logs(self, msgs: list):
        return msgs

    # ############### #
    # #### Tests #### #
    # ############### #

    def test_no_timeout(self):
        """Check that the plugin runs and returns successfully within the timeout."""
        loop = monitor.Monitor(
            DummySleepPlugin,
            {
                "server": self.server + "/null",
                "run_timeout": 3,
                "heartbeat_interval": 3,
                "max_timeouts_before_exit": 0,
            },
        )
        loop.run_loop(1)

        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_event: dict = r.json()[0]
        # standardise variable timestamps
        dt = datetime.datetime.now(datetime.timezone.utc)
        out_event["timestamp"] = dt.isoformat()
        self.assertEqual(
            out_event,
            dump(
                azm.StatusEvent(
                    model_version=azm.CURRENT_MODEL_VERSION,
                    kafka_key="runner-placeholder",
                    timestamp=dt,
                    author=azm.Author(
                        category="plugin",
                        name="DummySleepPlugin",
                        version="1.0",
                        security=None,
                    ),
                    entity=azm.StatusEvent.Entity(
                        input=azm.BinaryEvent(**out_event["entity"]["input"]),  # Don't care about this
                        status=State.Label.COMPLETED_EMPTY,
                        runtime=out_event["entity"]["runtime"],
                        results=[],
                    ),
                ),
            ),
        )
        print("------ACTUAL------")
        print(self.dummy_log_handler.logs)
        # Check the log messages, should be empty
        self.assertEqual(
            self._filter_error_from_logs(self.dummy_log_handler.logs),
            [],
        )

    @pytest.mark.timeout(10)
    def test_simple_timeout_and_catch_with_return(self):
        """Check that the plugin times out after 1 second, both with and without catching the exception."""
        # p1 just runs for 2 seconds and is aborted by the exception
        loop1 = monitor.Monitor(
            DummySleepPlugin,
            {
                "server": self.server + "/null",
                "run_timeout": 1,
                "heartbeat_interval": 1,
                "max_timeouts_before_exit": 0,
            },
        )
        # p2 catches the timeout exception and sleeps 2 more seconds before returning
        loop2 = monitor.Monitor(
            DummySleepPlugin,
            {
                "server": self.server + "/null",
                "run_timeout": 1,
                "heartbeat_interval": 1,
                "max_timeouts_before_exit": 0,
                "delay_after_exception": 2,
            },
        )

        for loop in (loop1, loop2):
            self.assertRaises(monitor.PluginTimeoutError, loop.run_loop, 1)

            r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
            r.raise_for_status()
            out_event: dict = r.json()[0]
            # standardise variable timestamps
            dt = datetime.datetime.now(datetime.timezone.utc)
            out_event["timestamp"] = dt.isoformat()
            self.assertEqual(
                out_event,
                dump(
                    azm.StatusEvent(
                        model_version=azm.CURRENT_MODEL_VERSION,
                        kafka_key="runner-placeholder",
                        timestamp=dt,
                        author=azm.Author(
                            category="plugin",
                            name="DummySleepPlugin",
                            version="1.0",
                            security=None,
                        ),
                        entity=azm.StatusEvent.Entity(
                            input=azm.BinaryEvent(**out_event["entity"]["input"]),  # Don't care about this
                            status=State.Label.ERROR_TIMEOUT,
                            error="PluginTimeout Error",
                            message="DummySleepPlugin timed out on job '1234'",
                            runtime=out_event["entity"]["runtime"],
                            results=[],
                        ),
                    ),
                ),
            )
        print("------ACTUAL------")
        print(self.dummy_log_handler.logs)
        # Check the log messages - a warning is expected since our timeout is less than the default heartbeat time
        self.assertEqual(
            self._filter_error_from_logs(self.dummy_log_handler.logs),
            [
                "WARNING: Execution of DummySleepPlugin timed out after 1 seconds for job with id '1234'",
                "ERROR: Exiting due to hitting limit of 0 timeouts",
                "WARNING: Execution of DummySleepPlugin timed out after 1 seconds for job with id '1234'",
                "ERROR: Exiting due to hitting limit of 0 timeouts",
            ],
        )

    @pytest.mark.timeout(10)
    def test_timeout_with_catch_no_return(self):
        """Tests that the expected errors occur when a plugin catches the timeout exception and doesn't terminate."""
        loop = monitor.Monitor(
            DummySleepPlugin,
            {
                "server": self.server + "/null",
                "run_timeout": 1,
                "heartbeat_interval": 1,
                "max_timeouts_before_exit": 0,
                "delay_after_exception": 5,
            },
        )
        self.assertRaises(
            monitor.PluginTimeoutError, loop.run_loop, 2
        )  # Set 2 jobs limit, but it should exit on error after the first one

        r = httpx.get("%s/mock/get_var/last_request_body" % self.server)
        r.raise_for_status()
        out_event: dict = r.json()[0]
        # standardise variable timestamps
        dt = datetime.datetime.now(datetime.timezone.utc)
        out_event["timestamp"] = dt.isoformat()
        self.assertEqual(
            out_event,
            dump(
                azm.StatusEvent(
                    model_version=azm.CURRENT_MODEL_VERSION,
                    kafka_key="runner-placeholder",
                    timestamp=dt,
                    author=azm.Author(
                        category="plugin",
                        name="DummySleepPlugin",
                        version="1.0",
                        security=None,
                    ),
                    entity=azm.StatusEvent.Entity(
                        input=azm.BinaryEvent(**out_event["entity"]["input"]),  # Don't care about this
                        status=State.Label.ERROR_TIMEOUT,
                        error="PluginTimeout Error",
                        message="DummySleepPlugin timed out on job '1234'",
                        runtime=out_event["entity"]["runtime"],
                        results=[],
                    ),
                ),
            ),
        )
        print("------ACTUAL------")
        print(self.dummy_log_handler.logs)
        # Check the log messages are as expected
        self.assertEqual(
            self._filter_error_from_logs(self.dummy_log_handler.logs),
            [
                "WARNING: Execution of DummySleepPlugin timed out after 1 seconds for job with id '1234'",
                "ERROR: Exiting due to hitting limit of 0 timeouts",
            ],
        )

    @pytest.mark.timeout(15)
    def test_timeout_count_limit(self):
        """Test that the runner exits after the specified number of timeouts."""
        loop = monitor.Monitor(
            DummySleepPlugin,
            {
                "server": self.server + "/null",
                "run_timeout": 1,
                "heartbeat_interval": 1,
                "max_timeouts_before_exit": 3,
            },
        )

        self.assertRaises(
            monitor.PluginTimeoutError, loop.run_loop, job_limit=10
        )  # Should stop after 3 when they all time out
        print("------ACTUAL------")
        print(self.dummy_log_handler.logs)
        # Check the log messages are as expected
        self.assertEqual(
            self._filter_error_from_logs(self.dummy_log_handler.logs),
            [
                "WARNING: Execution of DummySleepPlugin timed out after 1 seconds for job with id '1234'",
                "WARNING: Execution of DummySleepPlugin timed out after 1 seconds for job with id '1234'",
                "WARNING: Execution of DummySleepPlugin timed out after 1 seconds for job with id '1234'",
                "ERROR: Exiting due to hitting limit of 3 timeouts",
            ],
        )



from __future__ import annotations

import io
import time
import unittest
from typing import Any, ClassVar

import httpx
from azul_bedrock import dispatcher
from azul_bedrock import models_network as azm
from bitarray import bitarray

from azul_runner import local
from azul_runner.storage import StorageError, StorageProxyFile

from . import mock_dispatcher as md
from . import plugin_support as sup


class TestStorageProxyFile(unittest.TestCase):
    mock_server: ClassVar[md.MockDispatcher]
    server: ClassVar[str]  # Endpoint to the mock server in the form 'http://host:port'

    @classmethod
    def setUpClass(cls) -> None:
        cls.mock_server, cls.server = sup.setup_mock_dispatcher()
        cls.dp = dispatcher.DispatcherAPI(
            events_url=cls.server,
            data_url=cls.server,
            retry_count=2,
            timeout=2,
            author_name="TestStorageProxyFile",
            author_version="1",
            deployment_key="scorpion",
        )
        cls.editor = md.Editor(cls.server)
        cls.editor.set_stream("500-internal-server-error", 500, b"")
        cls.editor.set_stream("test-ff-128", 206, b"\xff" * 128)
        cls.editor.set_stream("test-ff-129", 206, b"\xff" * 129)
        cls.editor.set_stream("test-ff-256", 206, b"\xff" * 256)
        cls.editor.set_stream("test-ff-192", 206, b"\xff" * 192)
        cls.editor.set_stream("test-ff-191", 206, b"\xff" * 191)
        cls.editor.set_stream("test-zeros-256", 206, b"\x00" * 256)
        cls.editor.set_stream("test-zeros-255", 206, b"\x00" * 255)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.mock_server.stop()
        cls.mock_server.kill()

    # ############### #
    # #### Tests #### #
    # ############### #

    def test_instantiate(self):
        with StorageProxyFile("source", "content", "test-zeros-128", dp=self.dp) as f:
            # Size should now be None on first init, until something is read
            self.assertEqual(f._size, None)

    def test_invalid_params(self):
        with StorageProxyFile("source", "content", "test-zeros-256", dp=self.dp) as f:
            self.assertRaisesRegex(TypeError, "read length must be an int", f.read, 2.5)
            self.assertRaisesRegex(ValueError, "read length must be positive or -1", f.read, -50)
            self.assertRaisesRegex(ValueError, "start_chunk cannot be negative", f._request_chunks, -1, 20)
        with StorageProxyFile("source", "content", "INVALID URL", dp=self.dp) as f:
            self.assertRaisesRegex(StorageError, r"Got 404 requesting INVALID URL", f.read)
        with StorageProxyFile("source", "content", "nosuchfile", dp=self.dp) as f:
            self.assertRaisesRegex(StorageError, r"Got 404 requesting nosuchfile", f.read)

    def test_size_changed_assertion(self):
        # Must use a small chunk size for this one, or it will cache the whole file and not pick up the changed size
        with StorageProxyFile("source", "content", "test-zeros-256", chunk_size=16, dp=self.dp) as f:
            f.read(1)
            f._hash = "test-zeros-255"
            self.assertRaisesRegex(AssertionError, "Object size mismatch", f.read)
        with StorageProxyFile("source", "content", "test-zeros-256", chunk_size=256, dp=self.dp) as f:
            f.read(1)
            f._hash = "test-zeros-255"
            # No assertion error, as the entire file is already cached
            f.read()

    def test_internal_server_error(self):
        """Test remote internal server error."""
        with StorageProxyFile("source", "content", "test-zeros-256", dp=self.dp) as f:
            f._hash = "500-internal-server-error"
            self.assertRaisesRegex(StorageError, "Got 500 requesting 500-internal-server-error", f.read, 1)

    def test_chunkmap_and_cache(self):
        """Tests that the chunkmap bitarray behaves properly"""
        with StorageProxyFile("source", "content", "test-ff-128", chunk_size=32, dp=self.dp) as f:
            # Chunks should not yet be set, as nothing has been read / no requests to server
            self.assertEqual(f._chunks, bitarray())
            f.read(1)
            self.assertEqual(f._chunks, bitarray("1000"))
            f._content.seek(0)
            self.assertEqual(f._content.read(), b"\xff" * 32)
            f.seek(-1, io.SEEK_END)
            f.read(1)
            self.assertEqual(f._chunks, bitarray("1001"))
            f._content.seek(0)
            self.assertEqual(f._content.read(), b"\xff" * 32 + b"\x00" * 64 + b"\xff" * 32)
        with StorageProxyFile("source", "content", "test-ff-129", chunk_size=32, dp=self.dp) as f:
            f.MIN_LAST_CHUNK_SIZE = 0  # Disable last-chunk-merging for this test
            self.assertEqual(f._chunks, bitarray(""))
            f.seek(-1, io.SEEK_END)
            f.read(1)
            self.assertEqual(f._chunks, bitarray("00001"))
            f._content.seek(0)
            # Should actually have the last 32 bytes of real data, due to reading the final <chunk_size> bytes
            # (But only the last byte actually forms a valid chunk, since a byte in the 2nd-last chunk is missing)
            self.assertEqual(f._content.read(), b"\x00" * 97 + b"\xff" * 32)

    def test_last_chunk_merging(self):
        """Tests that auto-merging of small final chunks works as expected"""
        with StorageProxyFile("source", "content", "test-ff-192", chunk_size=128, dp=self.dp) as f:
            # Last chunk is >= the limit, so last chunk is fetched separately
            f.MIN_LAST_CHUNK_SIZE = 0.5
            self.assertEqual(f._chunks, bitarray())  # Should be zero-length since the size is unknown.
            f.read(1)
            self.assertEqual(f._chunks, bitarray("10"))
            f._content.seek(0)
            self.assertEqual(f._content.read(), b"\xff" * 128)
            f.seek(-1, io.SEEK_END)
            f.read(1)
            self.assertEqual(f._chunks, bitarray("11"))
            f._content.seek(0)
            self.assertEqual(f._content.read(), b"\xff" * 192)
        with StorageProxyFile("source", "content", "test-ff-191", chunk_size=128, dp=self.dp) as f:
            # The size is initially not known, so the extra partial chunk won't be read.
            f.MIN_LAST_CHUNK_SIZE = 0.5
            self.assertEqual(f._chunks, bitarray())
            f.read(1)
            self.assertEqual(f._chunks, bitarray("10"))
        with StorageProxyFile("source", "content", "test-ff-191", chunk_size=128, expected_size=191, dp=self.dp) as f:
            # Size known to be one byte smaller than the min size limit;  should auto-fetch the last chunk
            #  when the second-last chunk is read (in this case the first chunk)
            f.MIN_LAST_CHUNK_SIZE = 0.5
            self.assertEqual(f._chunks, bitarray("00"))
            f.read(1)
            self.assertEqual(f._chunks, bitarray("11"))
            f._content.seek(0)
            self.assertEqual(f._content.read(), b"\xff" * 191)

    def test_read_whole_file(self):
        with StorageProxyFile("source", "content", "test-zeros-256", chunk_size=16, dp=self.dp) as f:
            self.assertEqual(f.read(), b"\x00" * 256)
            self.assertEqual(self.editor.get_last_request().range, [0, 255])
            self.assertEqual(f.read(), b"")
            # Further requests should not go back to the server, as the full content has been acquired.
            f.seek(128)
            f.read()
            self.assertEqual(self.editor.get_last_request().range, [0, 255])

    def test_read_partial_chunks(self):
        with StorageProxyFile("source", "content", "test-zeros-256", chunk_size=16, dp=self.dp) as f:
            self.assertEqual(f.read(16), b"\x00" * 16)
            self.assertEqual(self.editor.get_last_request().range, [0, 15])

    def test_read_middle_chunk(self):
        with StorageProxyFile("source", "content", "test-zeros-256", chunk_size=256, dp=self.dp) as f:
            # This should read the entire file due to large chunk size
            f.seek(128)
            self.assertEqual(f.read(64), b"\x00" * 64)

            self.assertEqual(self.editor.get_last_request().range, [0, 255])
        with StorageProxyFile("source", "content", "test-zeros-256", chunk_size=16, dp=self.dp) as f:
            # This should read only what's requested
            f.seek(128)
            self.assertEqual(f.read(64), b"\x00" * 64)
            self.assertEqual(self.editor.get_last_request().range, [128, 191])
            # This should not go to the server as it's already at the EOF. (size is known)
            f.seek(256)
            self.assertEqual(f.read(), b"")
            self.assertEqual(self.editor.get_last_request().range, [128, 191])
            # This does go to the server as the file was seeked back one and that block is not cached yet:
            f.seek(-1, io.SEEK_CUR)
            self.assertEqual(f.read(), b"\x00")
            self.assertEqual(self.editor.get_last_request().range, [240, 255])

    def test_read_past_end(self):
        with StorageProxyFile("source", "content", "test-zeros-256", chunk_size=32, dp=self.dp) as f:
            # Size is known so only request what is needed in this version of StorageProxyFile
            self.assertEqual(f.read(500), b"\x00" * 256)
            self.assertEqual(self.editor.get_last_request().range, [0, 255])

    def test_seek_from_end(self):
        """Tests seeking back from end on a newly-opened file"""
        with StorageProxyFile("source", "content", "test-zeros-256", dp=self.dp) as f:
            f.seek(-1, io.SEEK_END)
            self.assertEqual(f.tell(), 255)

    def test_seek_past_end(self):
        """Tests seeking past the end of a file and trying to read"""
        with StorageProxyFile("source", "content", "test-zeros-256", chunk_size=32, dp=self.dp) as f:
            f.read(32)
            self.assertEqual(f._size, 256)  # Should now know this
            f.seek(512)
            self.assertEqual(f.read(), b"")
            # Should not have gone to the server for this, as it's past EOF
            self.assertEqual(self.editor.get_last_request().range, [0, 31])
        with StorageProxyFile("source", "content", "test-zeros-256", chunk_size=32, dp=self.dp) as f:
            f.seek(512)
            self.assertEqual(f.read(), b"")
            # This time request from the server because the size was not known before the request (and got no content)
            self.assertIsNone(self.editor.get_last_request().range)
            # Should have been given the size
            self.assertEqual(f._size, 256)

    def test_expected_size(self):
        """Tests that there is no GET issued when an expected size is provided"""
        with StorageProxyFile("source", "content", "test-zeros-256", chunk_size=128, dp=self.dp) as f:
            # No expected size
            f.seek(-1, io.SEEK_END)
            self.assertEqual(f.tell(), 255)
            # Should have fetched the last chunk when trying to determine the size
            self.assertEqual(f._chunks, bitarray("01"))
        with StorageProxyFile(
            "source", "content", "test-zeros-256", chunk_size=128, expected_size=256, dp=self.dp
        ) as f:
            f.seek(-1, io.SEEK_END)
            self.assertEqual(f.tell(), 255)
            # Should not have fetched anything, but should have initialised _chunks
            self.assertEqual(f._chunks, bitarray("00"))

    def test_splitting(self):
        """Tests that requests are split into multiple GETs when the cached data is already in chunks."""
        with StorageProxyFile("source", "content", "test-ff-256", chunk_size=16, expected_size=256, dp=self.dp) as f:
            f.seek(128)
            f.read(16)
            self.assertEqual(f._chunks, bitarray("0000000010000000"))
            f.seek(0)
            # Make sure the whole file is fetched.
            self.assertEqual(f.read(), b"\xff" * 256)
            self.assertEqual(f._chunks, bitarray("1111111111111111"))
            # ... but the request should have been split in two, so this will be the range of the second request.
            self.assertEqual(self.editor.get_last_request().range, [144, 255])

        with StorageProxyFile("source", "content", "test-ff-256", chunk_size=16, dp=self.dp) as f:
            f.seek(32)
            f.read(16)
            f.seek(128)
            f.read(64)
            # No expected_size, but size should be collected after the first GET
            self.assertEqual(f._chunks, bitarray("0010000011110000"))
            f.seek(0)

            # Monkeypatch to check the expected three requests are made
            getlist = []

            def _getchunks(start, end):
                bsize = f._chunk_size
                getlist.append((start * bsize, end * bsize - 1))
                f._content.seek(start * bsize)
                f._content.write(b"\xff" * (end - start) * bsize)
                f._chunks[start:end] = True

            f._request_chunks = _getchunks

            self.assertEqual(f.read(193), b"\xff" * 193)
            self.assertEqual(f._chunks, bitarray("1111111111111000"))
            self.assertEqual(getlist, [(0, 31), (48, 127), (192, 207)])

    def test_init_data(self):
        """Tests for correct behaviour when input is init_data=b'...'"""
        # This also confirms that no HTTP requests are made, or an error would be raised due to invalid "url" and
        #  no f._session value set.
        with StorageProxyFile(
            "source", "content", "foo", init_data=b"This is the file content to be used", dp=self.dp
        ) as f:
            self.assertEqual(f._size, 35)
            self.assertEqual(f.read(10), b"This is th")
            self.assertEqual(f.read(), b"e file content to be used")

    def test_get_hash(self):
        """Tests for correct behaviour when input is init_data=b'...'"""
        # This also confirms that no HTTP requests are made, or an error would be raised due to invalid "url" and
        #  no f._session value set.
        content = b"This is the file content to be used"
        fi = local.gen_api_content(io.BytesIO(content), azm.DataLabel.CONTENT)
        with StorageProxyFile("source", "content", "foo", init_data=content, file_info=fi) as f:
            self.assertEqual(f.get_hash(), "a2da980d45b8fef3281dd222bb45c6725d62fc8a8b3ffda8fabea8c5211cd85f")
        with StorageProxyFile("source", "content", "foo", init_data=content, file_info=fi, dp=self.dp) as f:
            self.assertEqual(f.get_hash(), "a2da980d45b8fef3281dd222bb45c6725d62fc8a8b3ffda8fabea8c5211cd85f")

    def test_tempfile(self):
        """Tests that tempfile works."""
        with StorageProxyFile("source", "content", "test-ff-256", chunk_size=16, expected_size=256, dp=self.dp) as spf:
            f = spf.get_tempfile()
            f.seek(128)
            self.assertEqual(f.read(16), b"\xff" * 16)
            f.seek(0)
            # Make sure whole file was fetched.
            self.assertEqual(f.read(), b"\xff" * 256)

    def test_filepath(self):
        """Tests that filepath works."""
        with StorageProxyFile("source", "content", "test-ff-256", chunk_size=16, expected_size=256, dp=self.dp) as spf:
            path = spf.get_filepath()
            with open(path, "rb") as f:
                f.seek(128)
                self.assertEqual(f.read(16), b"\xff" * 16)
                f.seek(0)
                # Make sure whole file was fetched.
                self.assertEqual(f.read(), b"\xff" * 256)



from __future__ import annotations

import unittest
import uuid
from typing import Any

from azul_bedrock import exceptions_bedrock as azbe
from azul_bedrock import models_network as azm

from azul_runner import (
    FV,
    Event,
    EventData,
    Feature,
    FeatureType,
    JobResult,
    State,
    TestPlugin,
    add_settings,
    settings,
)
from azul_runner.models import EventParent

from . import plugin_support as sup


class BaseInputContentNoFilter(sup.DummyPlugin):
    SETTINGS = add_settings(
        filter_data_types={azm.DataLabel.TEST: []},
    )


class BaseInputContentAssertionPluginV1(sup.DummyPluginMinimum):
    SETTINGS = add_settings(
        filter_data_types={azm.DataLabel.TEST: ["text/plain"]},
    )


class BaseInputContentAssertionPluginV2(sup.DummyPluginMinimum):
    SETTINGS = add_settings(
        filter_data_types={azm.DataLabel.TEST: ["text/plain", "unknown", "network/tcpdump"]},
    )


class BaseInputContentAssertionPluginV3(sup.DummyPluginMinimum):
    SETTINGS = add_settings(
        filter_data_types={azm.DataLabel.TEST: ["network/tcpdump"]},
    )


# New file format checks


class TestExecutionWrapperInputChecking(TestPlugin):
    PLUGIN_TO_TEST = sup.DummyPlugin

    def test_input_content_assertions(self):
        """Checks that filter_data_types is working correctly."""
        self.assertRaisesRegex(
            AssertionError,
            r"Provided file format \{\} does not meet plugin requirements \{'test': \['text/plain'\]\}\n"
            r"magic: \[\]\n"
            r"mime types: \[\]",
            self.do_execution,
            plugin_class=BaseInputContentAssertionPluginV1,
        )

        self.assertRaisesRegex(
            AssertionError,
            r"Provided file format \{'test': \{'unknown'\}, '\*': \{'unknown'\}\} does not meet plugin requirements \{'test': \['text/plain'\]\}\n"
            r"magic: \['data'\]\n"
            r"mime types: \['application/octet-stream'\]",
            self.do_execution,
            data_in=[(azm.DataLabel.TEST, b"\xff\xff\xe8\x00\xfd")],
            plugin_class=BaseInputContentAssertionPluginV1,
        )
        # Should not raise
        self.do_execution(data_in=[(azm.DataLabel.TEST, b"plain text data")])
        self.do_execution(verify_input_content=False)
        self.do_execution(data_in=[(azm.DataLabel.TEST, b"\xff\xff\xe8\x00\xfd")], verify_input_content=False)

        # Should not raise with no requirements
        self.do_execution(data_in=[(azm.DataLabel.TEST, b"plain text data")], plugin_class=BaseInputContentNoFilter)
        self.do_execution(
            data_in=[(azm.DataLabel.TEST, b"\xff\xff\xe8\x00\xfd")], plugin_class=BaseInputContentNoFilter
        )

        self.do_execution(
            data_in=[(azm.DataLabel.TEST, b"plain text data")], plugin_class=BaseInputContentAssertionPluginV2
        )
        # Should not raise
        self.do_execution(
            data_in=[(azm.DataLabel.TEST, b"plain text data")], plugin_class=BaseInputContentAssertionPluginV2
        )
        self.do_execution(verify_input_content=False, plugin_class=BaseInputContentAssertionPluginV2)
        self.do_execution(
            data_in=[(azm.DataLabel.TEST, b"\xff\xff\xe8\x00\xfd")],
            verify_input_content=False,
            plugin_class=BaseInputContentAssertionPluginV2,
        )

        self.assertRaisesRegex(
            AssertionError,
            r"Provided file format \{'test': \{'text/plain'\}, '\*': \{'text/plain'\}\} does not meet plugin requirements \{'test': \['network/tcpdump'\]\}\n"
            r"magic: \['ASCII text, with no line terminators'\]\n"
            r"mime types: \['text/plain'\]",
            self.do_execution,
            data_in=[(azm.DataLabel.TEST.value, b"plain text data")],
            plugin_class=BaseInputContentAssertionPluginV3,
        )


class TestEventLimits(TestPlugin):
    PLUGIN_TO_TEST = sup.DummyPlugin

    class DPTooLarge(sup.DummyPlugin):
        def execute(self, job) -> dict:
            tmp = "a" * (1024 * 1024 * 2 + 512)
            print(len(tmp))
            print(tmp[:20])
            self.add_info({"large": tmp})

    def test_too_large(self):
        """Limit total size of status message."""

        # Generate a dynamic class with input content requirements
        with self.assertRaises(azbe.NetworkDataException) as e:
            self.do_execution(plugin_class=self.DPTooLarge)

        self.assertIn("event produced by plugin was too large: ", e.exception.external)


class TestAlterConfig(TestPlugin):
    PLUGIN_TO_TEST = sup.DummyPlugin

    class DPTooLarge(sup.DummyPlugin):
        def execute(self, job) -> dict:
            tmp = "a" * (1024 * 1024 * 2 + 512)
            print(len(tmp))
            print(tmp[:20])
            self.add_info({"large": tmp})

        def _alter_config(self, config: settings.Settings) -> settings.Settings:
            config.filter_data_types = {"*": ["text/plain"]}
            return config

    def test_too_large(self):
        """Limit total size of status message."""

        # Generate a dynamic class with input content requirements
        p = self.DPTooLarge()
        self.assertEqual(p.cfg.filter_data_types, {"*": ["text/plain"]})


class TestRequireConfig(TestPlugin):
    """Verify that a plugin can enforce config requirements.
    The embedded tests will fail if PLUGIN_TO_TEST_CONFIG is not set.
    """

    class PluginRequireConfig(sup.DummyPlugin):
        SETTINGS = add_settings(custom_config=(str, ""))

        def __init__(self, config: settings.Settings | dict = None) -> None:
            super().__init__(config)
            if not self.cfg.custom_config:
                raise Exception("plugin requires custom_config property to be set")

    PLUGIN_TO_TEST = PluginRequireConfig
    PLUGIN_TO_TEST_CONFIG = {"custom_config": "yeah"}


class TestMultiStreamResults(TestPlugin):
    """
    Tests correct handling of per-stream results by the runner
    """

    class DP(sup.DummyPlugin):
        def execute(self, job) -> dict:
            e = self.get_data_event("nonexistent")
            e.add_feature_values("example_int", 42)

    PLUGIN_TO_TEST = DP

    def test_invalid_stream_error(self):
        """Tests that an error is raised when a result is returned for a non-existent input stream"""
        result = self.do_execution(
            data_in=[(azm.DataLabel.TEST, b"foo")],
        )
        self.assertRegex(result.state.message, "ResultError: results for non-existent stream")
        self.assertEqual(
            result.state,
            State(
                label=State.Label.ERROR_OUTPUT,
                failure_name="Invalid Plugin Output",
                message=result.state.message,
            ),
        )


class DummyMappingPlugin(sup.DummyPlugin):
    _IS_USING_PUSHER = True
    FEATURES = [Feature("value", "A return value", type=FeatureType.Integer)]

    def execute(self, job):
        self.add_feature_values("value", 20)


class TestPluginMappedOutput(TestPlugin):
    """Test custom output action types doesn't negatively effect plugin output."""

    PLUGIN_TO_TEST = DummyMappingPlugin

    def test_mapped_output(self):
        result = self.do_execution()
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[Event(sha256="test_entity", features={"value": [FV("20")]})],
            ),
        )


class TestPluginOutputStream(TestPlugin):
    """
    Tests for the handling of add_entity_content
    """

    PLUGIN_TO_TEST = sup.DummyPlugin

    class DPAddAltStream(sup.DummyPlugin):
        def execute(self, job):
            self.add_data("text", {"language": "value"}, b"Text data stream")

    def test_add_alt_stream(self):
        # Test a plugin successfully adding an additional stream result
        result = self.do_execution(plugin_class=self.DPAddAltStream)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="test_entity",
                        data=[
                            EventData(
                                hash="903e0722b8ebdfc5547af8ca1a9c8a76d0a7b067a7b1a561dfff13593e76d5a8",
                                label=azm.DataLabel.TEXT,
                                language="value",
                            )
                        ],
                    )
                ],
                data={"903e0722b8ebdfc5547af8ca1a9c8a76d0a7b067a7b1a561dfff13593e76d5a8": b"Text data stream"},
            ),
            inspect_data=True,
        )

    class DPAdditionCommonFeatureMalformed(sup.DummyPlugin):
        def execute(self, job):
            return self.is_malformed("This file is seriously malformed")

    def test_addition_common_feature_malformed(self):
        result = self.do_execution(plugin_class=self.DPAdditionCommonFeatureMalformed)
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED_WITH_ERRORS, message="This file is seriously malformed"),
                events=[
                    Event(
                        sha256="test_entity",
                        features={"malformed": [FV("This file is seriously malformed")]},
                    ),
                ],
            ),
        )

    class DPMutatingAugStream(sup.DummyPlugin):
        """Plugin that generates random sha256's"""

        def execute(self, job):
            unique_id_1 = str(uuid.uuid4())
            self.add_data(azm.DataLabel.TEST, {}, unique_id_1.encode())

    def test_mutating_augmented_stream(self):
        """Test that the test case fails if you mutate the augmented streams.

        Important to limit data produced by plugins.
        """

        # Verify assertion error is raised with inconsistent outputs
        with self.assertRaises(AssertionError):
            self.do_execution(
                plugin_class=self.DPMutatingAugStream,
            )

        # Verify bypassing the check is possible.
        result = self.do_execution(
            plugin_class=self.DPMutatingAugStream,
            check_consistent_augmented_stream=False,
        )
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[Event(sha256="test_entity", data=[EventData(hash="0", label="test")])],
                data={"0": b""},
            ),
            strip_hash=True,
        )

    class DPMutatingChildStream(sup.DummyPlugin):
        """Plugin that generates random sha256's"""

        def execute(self, job):
            unique_id_1 = str(uuid.uuid4())
            self.add_child_with_data({"action": "extracted"}, unique_id_1.encode())

    def test_mutating_child_stream(self):
        """Test that the test case fails if you mutate the child streams.

        Important to limit data produced by plugins.
        """
        # Verify assertion error is raised with inconsistent outputs
        with self.assertRaises(AssertionError):
            self.do_execution(
                plugin_class=self.DPMutatingChildStream,
            )

        # Verify bypassing the check is possible.
        result = self.do_execution(
            plugin_class=self.DPMutatingChildStream,
            check_consistent_augmented_stream=False,
        )
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(
                        sha256="0",
                        parent=EventParent(sha256="test_entity"),
                        relationship={"action": "extracted"},
                        data=[EventData(hash="0", label="content")],
                    )
                ],
                data={"0": b""},
            ),
            strip_hash=True,
        )

    class DPMutatingMultipluginStream(sup.DummyPlugin):
        """Plugin that generates random sha256's"""

        def __init__(self, config: dict[str, dict[str, Any]] = None) -> None:
            super().__init__(config)
            self._multiplugins = {}  # remove parent plugins
            self.register_multiplugin(None, None, lambda x: self.execute(x))
            self.register_multiplugin(None, None, lambda x: self.execute(x))

        def execute(self, job):
            unique_id_1 = str(uuid.uuid4())
            self.add_child_with_data({"action": "extracted"}, unique_id_1.encode())
            unique_id_2 = str(uuid.uuid4())
            self.add_data(azm.DataLabel.TEST, {}, unique_id_2.encode())

    def test_mutating_multiplugin_stream(self):
        """Test that the test case fails if you mutate the augmented/child streams with a multi-plugin.

        Important to limit data produced by plugins.
        """
        # Verify assertion error is raised with inconsistent outputs
        with self.assertRaises(AssertionError):
            self.do_execution(
                plugin_class=self.DPMutatingMultipluginStream,
            )

        # Verify bypassing the check is possible.
        result = self.do_execution(
            plugin_class=self.DPMutatingMultipluginStream,
            check_consistent_augmented_stream=False,
        )
        self.assertJobResult(
            result,
            JobResult(
                state=State(State.Label.COMPLETED),
                events=[
                    Event(sha256="test_entity", data=[EventData(hash="1", label="test")]),
                    Event(
                        sha256="0",
                        parent=EventParent(sha256="test_entity"),
                        relationship={"action": "extracted"},
                        data=[EventData(hash="0", label="content")],
                    ),
                ],
                data={"0": b"", "1": b""},
            ),
            strip_hash=True,
        )


if __name__ == "__main__":
    unittest.main()



# flake8: noqa - Ruff's linting rules doesn't detect these exported functions and classes properly.
import contextlib

from azul_bedrock.models_network import DataLabel, FeatureType
from azul_bedrock.models_network import FeatureValue as APIFeatureValue

from .binary_plugin import BinaryPlugin
from .main import cmdline_run
from .models import (
    FV,
    Event,
    EventData,
    EventParent,
    Feature,
    FeatureValue,
    Filepath,
    Job,
    JobResult,
    State,
    Uri,
)
from .plugin import Plugin
from .settings import add_settings
from .storage import DATA_HASH, DATA_HASH_NAME, StorageProxyFile

append_all: list[str] = []
with contextlib.suppress(ImportError):
    from .test_utils import test_template
    from .test_utils.test_template import TestPlugin

    print("You have the azul-runner[test_utils] installed, this should only be used for development or testing.")
    append_all = ["TestPlugin", "test_template"]

EXPORTS = append_all + [
    "add_settings",
    "APIFeatureValue",
    "BinaryPlugin",
    "cmdline_run",
    "DATA_HASH_NAME",
    "DATA_HASH",
    "DataLabel",
    "Event",
    "EventData",
    "EventParent",
    "Feature",
    "FeatureType",
    "FeatureValue",
    "Filepath",
    "FV",
    "Job",
    "JobResult",
    "Plugin",
    "State",
    "StorageProxyFile",
    "Uri",
]

__all__ = list(EXPORTS)



"""Plugin settings, using pydantic environment parsing."""

from enum import Enum
from inspect import isclass
from typing import Any

import pydantic
from pydantic_settings import BaseSettings, SettingsConfigDict


class PluginError(Exception):
    """Generic error with plugin."""

    pass


class SetupError(PluginError):
    """Error relating to configuration/initialisation of plugin."""

    pass


_prefix = "plugin_"


class WatchTypeEnum(str, Enum):
    """Different file watch types."""

    PLAIN = ""
    GIT = "git"


class RunnerSettings(BaseSettings):
    """Runner specific environment variables that must not be altered by plugin implementations.

    This is used before a plugin is created, so cannot be shared with regular plugin settings.
    """

    model_config = SettingsConfigDict(env_prefix="runner_", extra="allow")

    # log level of runner + plugin loggers
    log_level: str = "WARNING"


class PluginBaseSettings(BaseSettings):
    """Base settings with prefix configuration already included."""

    model_config = SettingsConfigDict(env_prefix=_prefix, extra="ignore")


class Settings(BaseSettings):
    """Plugin specific environment variables parsed into settings object."""

    model_config = SettingsConfigDict(env_prefix=_prefix, extra="allow")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.heartbeat_interval < 1:
            raise SetupError(f"{self.heartbeat_interval=} too small")

        # Check that heartbeat and timeout values are sane
        if self.run_timeout and self.heartbeat_interval > self.run_timeout:
            raise SetupError(f"{self.heartbeat_interval=} greater than {self.run_timeout=}")

        if self.plugin_depth_limit <= 0:
            raise SetupError(f"{self.plugin_depth_limit=} must be positive")

        if self.server:
            if self.events_url or self.data_url:
                raise SetupError("cannot combine server with events_url or data_url")
            self.data_url = self.events_url = self.server

    # Seconds between heartbeat status messages.
    heartbeat_interval: int = 30
    # Seconds a plugin can run on a single sample before being aborted, 0 to never time out.
    # Set to 10 minutes by default so rare long running binaries are more likely to be processed.
    run_timeout: int = 600
    # Exit run loop after this many timeouts.
    # Useful in case a plugin doesn't clean up all resources when forcefully terminated.
    max_timeouts_before_exit: int = 100

    # Base URL to the dispatcher retrieve events and post results. eg "https://some.server/path".
    server: str = ""
    # dispatcher to use for event interaction
    events_url: str = ""
    # dispatcher to use for file data interaction
    data_url: str = ""
    # Seconds to wait before timing out API requests.
    request_timeout: int = 15
    # Times to retry API requests before worker dies.
    request_retry_count: int = 3
    # Content metadata entries to carry forward between jobs.
    # 0 = clear cache after every run.
    content_meta_cache_limit: int = 0
    # Hard limit of values for a single feature.
    max_values_per_feature: int = 1000
    # Hard limit of feature value length (opensearch limit is 32766)
    # Plugins should use text streams instead if values are sufficiently large
    max_value_length: int = 4000
    # Max event processing depth before auto opting out, -1 to disable.
    plugin_depth_limit: int = 10
    # Seconds to sleep if plugin was not ready to receive events.
    not_ready_backoff: int = 5
    # Prefix to be removed from Plugin Name at startup
    name_remove_prefix: str = "AzulPlugin"
    # Suffix to append to name on startup.
    name_suffix: str = ""
    # Suffix to append to version on startup.
    version_suffix: str = ""
    # Override plugin SECURITY with this on startup.
    security_override: str = ""
    # true if plugin code assumes that all streams can be pulled from s3.
    # false if plugin will handle 'FileNotFound' exceptions when performing .get_all_data() or similar.
    assume_streams_available: bool = False
    # a unique key for all instances of this plugin, typically matching the
    # the name of the parent deployment.
    deployment_key: str = ""

    # File watch, for running plugin with up-to-date files
    # if files in this path are modified, the plugin will reload
    watch_path: str = ""
    # if watch_type==git, then on plugin restart, will set last commit hash of watch_path as version_suffix
    watch_type: WatchTypeEnum = WatchTypeEnum.PLAIN
    # wait x seconds after first change event before restarting plugin
    watch_wait: int = 10

    # Memory limits
    # Enable the memory limiting functionality (disable by default because only works in Kubernetes).
    enable_mem_limits: bool = False
    # Fractional amount of max memory usage when a warning should be raise (e.g 0.8 == 80%)
    used_mem_warning_frac: float = 0.8
    # Fractional amount of max memory usage when plugin should be shutdown and the current job listed as an error.
    used_mem_force_exit_frac: float = 0.9
    # Max memory file location. (shouldn't need changing but just in case.)
    max_mem_file_path: str = "/sys/fs/cgroup/memory.max"
    # Current memory file location. (shouldn't need changing but just in case.)
    cur_mem_file_path: str = "/sys/fs/cgroup/memory.current"
    # Current memory summary file location. (shouldn't need changing but just in case.)
    cur_mem_summary_file_path: str = "/sys/fs/cgroup/memory.stat"
    # How often to check if the process is out of memory (milliseconds).
    mem_poll_frequency_milliseconds: int = 1000

    #
    # these options affect what events your plugin will receive to process
    #

    # process user submitted binaries at a high priority
    require_expedite: bool = True
    # process events as they arrive (i.e. only run on new binaries)
    require_live: bool = True
    # process events from the past (i.e. if version changes and you want to rerun plugin)
    require_historic: bool = True
    # only keep events that have 'content' stream below this size (must be greater than filter_min_content_size)
    # '0' indicates no restriction.
    # default to 200MiB (assume plugins will load and process whole file in memory)
    filter_max_content_size: pydantic.ByteSize = "200MiB"
    # only keep events that have 'content' stream above this size (must be less than max)
    # '0' indicates no restriction.
    filter_min_content_size: pydantic.ByteSize = 0
    # allow only specified event types
    filter_allow_event_types: list[str] = []
    # filter out events published by this plugin
    filter_self: bool = False
    # Require at least one data stream with label matching each key.
    # Outer key is the stream label. Event must have a stream with every label mentioned.
    # Inner list is all accepted file types. Event must have at least one matching file type.
    # Usually you will only want to filter with the content stream label and a limited list of file types.
    # See identify.py for valid file types.
    # Examples:
    # requires a content stream of any file type.
    # {'content': []}
    # requires a content stream either win32 or dos executable.
    # {'content': ["executable/windows/pe32", "executable/windows/dos"]}
    # requires any stream either win32 or dos executable.
    # {'*': ["executable/windows/pe32", "executable/windows/dos"]}
    # requires any stream with a dos exe and a blob stream as either gzip or bzip.
    # {'*': ['executable/windows/dos'], 'blob': ['archive/gzip', 'archive/bzip2']}
    filter_data_types: dict[str, list[str]] = {}

    # Number of concurrent instances, useful for plugins that wait a long time for an external (e.g Cape)
    # Launches this many subprocesses, each running their own instance of the plugin.
    concurrent_plugin_instances: int = 1


def add_settings(**field_definitions: Any) -> type[PluginBaseSettings]:
    """Create a pydantic settings class useable for plugins SETTINGS field.

    If overriding settings from runner, only the value needs to be supplied, not the type.

    Otherwise kwargs provided to this method are same as field_definitions from the pydantic docs:
        https://docs.pydantic.dev/2.6/api/base_model/#pydantic.create_model

    which are in the form:
        <name>=(<type>, <default>)
    e.g:
        timeout=(int, 300),
        default_passwords=(list[str], ["password1", "infected", "password"]),
        useful=(bool, True)
    """
    # Handle case where a plugin overrides existing config defined in Settings
    filled_fields = {}
    for k, v in field_definitions.items():
        if not isinstance(v, tuple):
            if k in Settings.model_fields:
                filled_fields[k] = (Settings.model_fields[k].annotation, v)
            else:
                raise Exception(f"A custom setting did not provide type information: '{k}'")
        else:
            filled_fields[k] = v

    return pydantic.create_model("PluginSettings", __base__=PluginBaseSettings, **filled_fields)


def parse_config(cls, in_cfg: dict) -> Settings:
    """Validate and transform/merge the plugin configs into unified config.

    cls is the class for the plugin
    """
    inheritance = cls.__mro__
    s = Settings()
    c = s.model_dump()

    # update with any configs passed in to __init__
    c.update(in_cfg)

    # plugins can override standard options and add custom config options
    # Note that it only looks at the plugin class and parent classes,
    #  so SETTINGS set on an *instance* will be ignored. (You shouldn't be doing that anyway.)
    for cls in reversed(inheritance):
        # check for config
        config = getattr(cls, "SETTINGS", None)
        if config is None:
            continue
        error_message = (
            f"Plugin misconfigured: SETTINGS is a type '{type(config)}' and value '{config}' but must use the"
            + " method 'azul_runner.add_settings()' to add config."
        )
        if not isclass(config) or not issubclass(config, PluginBaseSettings):
            raise SetupError(error_message)

        # Get custom configuration options for the plugin from the commandline, environment variables and defaults.
        # NOTE: ignores in_cfg found in Settings that aren't in the PluginSettings
        config_model = config(**in_cfg)
        if config_model.model_config.get("extra", None) != "ignore":
            raise SetupError("Bad Extra's parameter - " + error_message)

        # Dump the plugin configuration over top of current config
        c.update(config_model.model_dump())

    # sanity check / normalise
    return Settings(**c)



from __future__ import annotations

import datetime
import io
import unittest

from azul_bedrock import models_network as azm

from azul_runner import (
    BinaryPlugin,
    Feature,
    FeatureType,
    State,
    StorageProxyFile,
    add_settings,
    local,
    plugin_executor,
)
from azul_runner.models import Job
from azul_runner.storage import ProxyFileNotFoundError


class PluginNoData(BinaryPlugin):
    """
    Test class that will register successfully but has no execute method.
    """

    FEATURES = [
        Feature("sample_feature", "An output feature for the test plugin", type=FeatureType.String),
    ]
    SETTINGS = add_settings(assume_streams_available=False)

    def execute(self, job: Job) -> dict:
        digest = job.event.entity.sha256
        self.add_feature_values("sample_feature", digest)


class TestBasePluginStatic(unittest.TestCase):
    """Static test cases for binary template."""

    def test_generate_filter_string(self):
        """Test that the size limit filter string is generated correctly"""

        class P(BinaryPlugin):
            SETTINGS = add_settings(
                filter_data_types={"example": ["content-type-1", "content-type-2"]},
                filter_max_content_size=500,
            )

        p = P()
        self.assertEqual(p.cfg.filter_max_content_size, 500)
        self.assertEqual(p.cfg.filter_data_types, {"example": ["content-type-1", "content-type-2"]})

    def test_generate_filter_string_cmdline_config(self):
        """Test that the size limit filter string is generated correctly when passed via cmdline."""

        class P(BinaryPlugin):
            SETTINGS = add_settings(
                filter_data_types={"example": ["content-type-1", "content-type-2"]},
                filter_max_content_size=100,
            )

        p = P({"filter_max_content_size": "500"})
        self.assertEqual(p.cfg.filter_max_content_size, 500)
        self.assertEqual(p.cfg.filter_data_types, {"example": ["content-type-1", "content-type-2"]})

    def test_handle_no_data(self):
        """Test that the template correctly handles binaries with data unavailable"""

        # Test with a plugin that expects data
        class Test404(BinaryPlugin):
            def execute(self, job) -> dict:
                data = job.get_data()
                print(data)
                # Mimic the exception raised by StorageProxyFile on a 404
                raise ProxyFileNotFoundError(2, "Got 404 requesting (nothing)")

        p = Test404()

        streams = [
            StorageProxyFile(
                source="local",
                label=azm.DataLabel.CONTENT,
                hash="test_data",
                init_data=b"test",
                file_info=local.gen_api_content(
                    io.BytesIO(b"test"),
                ),
            )
        ]

        event = azm.BinaryEvent(
            model_version=azm.CURRENT_MODEL_VERSION,
            kafka_key="runner-placeholder",
            action=azm.BinaryAction.Enriched,
            timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
            source=azm.Source(
                name="source",
                path=[
                    azm.PathNode(
                        author=azm.Author(name="foo"),
                        action=azm.BinaryAction.Enriched,
                        timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                        sha256="test_entity",
                    ),
                ],
                timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
            ),
            author=azm.Author(name="TestServer", category="blah"),
            entity=azm.BinaryEvent.Entity(sha256="test_entity", datastreams=[], features=[], info={}),
        )
        job = Job(event=event)
        job.load_streams(local=streams)
        res = plugin_executor.run_plugin_with_job(p, job, None)
        self.assertEqual(
            res.state,
            State(State.Label.OPT_OUT, message="Plugin requires binary content but none exists"),
        )
        # Test that it still raises if the plugin doesn't require data
        # Test404.REQUIRES_DATA = False
        p.cfg.assume_streams_available = False
        self.assertRaisesRegex(ProxyFileNotFoundError, "Got 404 requesting", p.execute, job)

        # validate you can test plugins that don't need data
        self.assertEqual(PluginNoData().cfg.assume_streams_available, False)
        p = PluginNoData()
        entity = azm.BinaryEvent.Entity(sha256="test_entity", datastreams=[], features=[], info={})
        event = azm.BinaryEvent(
            model_version=azm.CURRENT_MODEL_VERSION,
            kafka_key="runner-placeholder",
            action=azm.BinaryAction.Enriched,
            timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
            source=azm.Source(
                name="source",
                path=[
                    azm.PathNode(
                        author=azm.Author(name="foo"),
                        action=azm.BinaryAction.Enriched,
                        timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
                        sha256="test_entity",
                    ),
                ],
                timestamp=datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc),
            ),
            author=azm.Author(name="TestServer", category="blah"),
            entity=entity,
        )
        job = Job(event=event)
        res = plugin_executor.run_plugin_with_job(p, job, None)
        self.assertEqual(res.state, State())

        # Test that it treats regular FileNotFound correctly
        # Test with a plugin that expects data
        class TestFNF(BinaryPlugin):
            def execute(self, entity):
                data = job.get_data()
                with open(".nonexistent_file_3284767538574862745", "rb"):
                    raise AssertionError("Didn't get a 'file not found' ...")
                return

        self.assertEqual(TestFNF().cfg.assume_streams_available, True)
        p = TestFNF()
        self.assertRaises(FileNotFoundError, p.execute, job)



"""Guarantee a valid path to a SpooledTemporaryFile after rollover has occurred."""

import tempfile


class SpooledNamedTemporaryFile(tempfile.SpooledTemporaryFile):
    """Modified tempfile.SpooledTemporaryFile() to support file names."""

    def rollover(self):
        """Overwrite rollover to use NamedTemporaryFile."""
        if self._rolled:
            return
        file = self._file
        newfile = self._file = tempfile.NamedTemporaryFile(**self._TemporaryFileArgs)
        del self._TemporaryFileArgs

        pos = file.tell()
        if hasattr(newfile, "buffer"):
            newfile.buffer.write(file.detach().getvalue())
        else:
            newfile.write(file.getvalue())
        newfile.seek(pos, 0)

        self._rolled = True
